#!/usr/bin/env python3

"""
KISS serial interface handler.  This allows basic support for talking to
KISS-based TNCs, managing the byte stuffing/unstuffing.
"""

from enum import Enum
from asyncio import Protocol, ensure_future, get_event_loop
from serial_asyncio import create_serial_connection
from serial import EIGHTBITS, PARITY_NONE, STOPBITS_ONE
from .signal import Signal
from binascii import b2a_hex
import time
import logging


# Constants


BYTE_FEND   = 0xc0
BYTE_FESC   = 0xdb
BYTE_TFEND  = 0xdc
BYTE_TFESC  = 0xdd

CMD_DATA    = 0x00
CMD_TXDELAY = 0x01
CMD_P       = 0x02
CMD_SLOTTIME= 0x03
CMD_TXTAIL  = 0x04
CMD_FDUPLEX = 0x05
CMD_SETHW   = 0x06
CMD_RETURN  = 0x0f


# States


class KISSDeviceState(Enum):
    """
    States permitted by a KISS device:
    - CLOSED: Serial port is closed
    - INIT: Serial port just opened, TNC may be in TNC2-mode and
            the library is attempting to put it into KISS mode.
    - OPEN: Serial port is open, TNC in KISS mode.
    - CLOSING: Close instruction just received.  Putting TNC back into
      TNC2-mode if requested then closing the port.
    """
    CLOSED = 0
    OPENING = 1
    OPEN = 2
    CLOSING = 3


# Command classes


class KISSCommand(object):
    """
    KISS base command class
    """

    # Known commands: this will be populated later.
    _KNOWN_COMMANDS = {}

    @classmethod
    def _register(cls, cmd, subclass):
        assert cmd not in cls._KNOWN_COMMANDS
        cls._KNOWN_COMMANDS[cmd] = subclass

    @classmethod
    def _stuff_bytes(cls, data):
        """
        Byte-stuff incoming byte string.
        """
        for byte in data:
            if byte == BYTE_FEND:
                yield BYTE_FESC
                yield BYTE_TFEND
            elif byte == BYTE_FESC:
                yield BYTE_FESC
                yield BYTE_TFESC
            else:
                yield byte

    @classmethod
    def _unstuff_bytes(cls, data):
        """
        Un-byte-stuff incoming byte string.
        """
        last = None
        for byte in data:
            if byte == BYTE_FESC:
                if last == BYTE_FESC:
                    yield last
                else:
                    last = BYTE_FESC
            elif last == BYTE_FESC:
                if byte == BYTE_TFEND:
                    yield BYTE_FEND
                elif byte == BYTE_TFESC:
                    yield BYTE_FESC
                else:
                    yield last
                    yield byte
                last = None
            else:
                yield byte

    @classmethod
    def decode(cls, frame):
        """
        Decode a raw KISS frame.
        """
        frame = bytearray(cls._unstuff_bytes(frame))
        port = frame[0] >> 4
        cmd  = frame[0] & 0x0f
        subclass = cls._KNOWN_COMMANDS.get(cmd, cls)
        return subclass(port=port, cmd=cmd, payload=frame[1:])

    def __init__(self, port, cmd, payload=None):
        self._port = port
        self._cmd = cmd
        self._payload = payload

    def __bytes__(self):
        out = bytearray([
            ((self._port & 0x0f) << 4) | (self._cmd & 0x0f)
        ])
        if self._payload:
            out += self._payload

        # Encode the byte sequences
        return bytes(self._stuff_bytes(out))

    def __str__(self):
        return '%s{Port %d, Cmd 0x%02x, Payload %s}' % (
                self.__class__.__name__,
                self.port,
                self.cmd,
                b2a_hex(self.payload).decode()
        )

    @property
    def port(self):
        return self._port

    @property
    def cmd(self):
        return self._cmd

    @property
    def payload(self):
        return self._payload


class KISSCmdReturn(KISSCommand):
    """
    Emit a return command to the TNC.
    """
    def __init__(self):
        super(KISSCmdReturn, self).__init__(port=15, cmd=CMD_RETURN)
KISSCommand._register(CMD_RETURN, KISSCmdReturn)


class KISSCmdData(KISSCommand):
    def __init__(self, port, payload, cmd=CMD_DATA):
        assert cmd == CMD_DATA
        super(KISSCmdData, self).__init__(port=port,
                cmd=CMD_DATA, payload=payload)
KISSCommand._register(CMD_DATA, KISSCmdData)


# KISS device interface


class BaseKISSDevice(object):
    """
    Base class for a KISS device.  This may have between 1 and 16 KISS
    ports hanging off it.
    """
    def __init__(self, reset_on_close=True,
            send_block_size=128, send_block_delay=0.1,
            kiss_commands=['INT KISS', 'RESET'],
            log=None, loop=None, **kwargs):
        if log is None:
            log = logging.getLogger(self.__class__.__module__)
        if loop is None:
            loop = get_event_loop()
        self._log = log
        self._protocol = None
        self._rx_buffer = bytearray()
        self._tx_buffer = bytearray()
        self._loop = loop
        self._port = {}
        self._state = KISSDeviceState.CLOSED
        self._open_time = 0
        self._reset_on_close = reset_on_close
        self._kiss_commands = kiss_commands
        self._kiss_rem_commands = None
        self._frame_seen = False
        self._send_block_size = send_block_size
        self._send_block_delay = send_block_delay

    def _receive(self, data):
        """
        Handle incoming data by appending to our receive buffer.  The
        data given may contain partial frames.
        """
        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('RECV RAW %r', b2a_hex(data).decode())

        self._rx_buffer += data
        if self._state == KISSDeviceState.OPENING:
            self._loop.call_soon(self._check_open)
        else:
            self._loop.call_soon(self._receive_frame)

    def _send(self, frame):
        """
        Send a frame via the underlying transport.
        """
        rawframe = bytes(frame)

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('XMIT FRAME %r', b2a_hex(rawframe).decode())

        if not self._tx_buffer.endswith(bytearray([BYTE_FEND])):
            self._tx_buffer += bytearray([BYTE_FEND])

        self._tx_buffer += bytes(rawframe) \
                         + bytearray([BYTE_FEND])
        self._loop.call_soon(self._send_data)

    def _receive_frame(self):
        """
        Scan the receive buffer for incoming frames and send these to the
        underlying device.  If more than one frame is present, schedule
        ourselves again with the IO loop.
        """
        # Locate the first FEND byte
        try:
            start = self._rx_buffer.index(BYTE_FEND)
        except ValueError:
            # No frames waiting
            self._rx_buffer = bytearray()
            return

        self._log.debug('RECV FRAME start at %d', start)

        # Discard the proceeding junk
        self._rx_buffer = self._rx_buffer[start:]
        del start
        assert self._rx_buffer[0] == BYTE_FEND

        # Locate the last FEND byte of the frame
        try:
            end = self._rx_buffer.index(BYTE_FEND, 1)
        except ValueError:
            # Uhh huh, so frame is incomplete.
            return

        self._log.debug('RECV FRAME end at %d', end)

        # Everything between those points is our frame.
        frame = self._rx_buffer[1:end]
        self._rx_buffer = self._rx_buffer[end:]

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('RECEIVED FRAME %s, REMAINING %s',
                    b2a_hex(frame).decode(),
                    b2a_hex(self._rx_buffer).decode()
            )

        # Two consecutive FEND bytes are valid, ignore these "empty" frames
        if len(frame) > 0:
            # Decode the frame
            self._loop.call_soon(self._dispatch_rx_frame,
                    KISSCommand.decode(frame))

        # If we just have a FEND, stop here.
        if bytes(self._rx_buffer) == bytearray([BYTE_FEND]):
            return

        # If there is more to send, call ourselves via the IO loop
        if len(self._rx_buffer):
            self._loop.call_soon(self._receive_frame)

    def _dispatch_rx_frame(self, frame):
        """
        Pass a frame to the underlying interface.
        """
        try:
            port = self._port[frame.port]
        except KeyError:
            # Port not defined.
            self._log.debug('RECV FRAME dropped %s', frame)
            return

        # Dispatch this frame to the port.  Swallow exceptions so we
        # don't choke the IO loop.
        self._log.debug('RECV FRAME dispatch %s', frame)
        try:
            port._receive_frame(frame)
        except:
            logging.getLogger(self.__class__.__module__).exception(
                    'Port %s failed to handle frame %s',
                    port, frame
            )

    def _send_data(self):
        """
        Send the next block of data waiting in the buffer.
        """
        data = self._tx_buffer[:self._send_block_size]
        self._tx_buffer = self._tx_buffer[self._send_block_size:]

        if self._log.isEnabledFor(logging.DEBUG):
            self._log.debug('XMIT RAW %r', b2a_hex(data).decode())

        self._send_raw_data(data)

        # If we are closing, wait for this to be sent
        if (self._state == KISSDeviceState.CLOSING) and \
                (len(self._tx_buffer) == 0):
            self._close()
            return

        if self._tx_buffer:
            self._loop.call_later(self._send_block_delay, self._send_data)

    def _init_kiss(self):
        assert self.state == KISSDeviceState.OPENING, \
                'Device is not opening'

        self._kiss_rem_commands = self._kiss_commands.copy()
        self._send_kiss_cmd()

    def _send_kiss_cmd(self):
        try:
            command = self._kiss_rem_commands.pop(0)
        except IndexError:
            # Should be open now.
            self._open_time = time.time()
            self._state = KISSDeviceState.OPEN
            self._rx_buffer = bytearray()
            return

        self._log.debug('Sending %r', command)
        command = command.encode('US-ASCII')
        self._rx_buffer = bytearray()
        for bv in command:
            self._send_raw_data(bytes([bv]))
            time.sleep(0.1)
        self._send_raw_data(b'\r')
        self._loop.call_later(0.5, self._check_open)

    def _check_open(self):
        """
        Handle opening of the port
        """
        self._loop.call_soon(self._send_kiss_cmd)

    def __getitem__(self, port):
        """
        Retrieve an instance of a specified port.
        """
        try:
            return self._port[port]
        except KeyError:
            pass

        self._log.debug('OPEN new port %d', port)
        p = KISSPort(self, port, log=self._log.getChild('port%d' % port))
        self._port[port] = p
        return p

    @property
    def state(self):
        return self._state

    def open(self):
        assert self.state == KISSDeviceState.CLOSED, \
                'Device is not closed'
        self._log.debug('Opening device')
        self._state = KISSDeviceState.OPENING
        self._open()

    def close(self):
        assert self.state == KISSDeviceState.OPEN, \
                'Device is not open'
        self._log.debug('Closing device')
        self._state = KISSDeviceState.CLOSING
        if self._reset_on_close:
            self._send(KISSCmdReturn())
        else:
            self._close()


class BaseTransportDevice(BaseKISSDevice):
    def __init__(self, *args, **kwargs):
        super(BaseTransportDevice, self).__init__(*args, **kwargs)
        self._transport = None

    def _make_protocol(self):
        """
        Return a Protocol instance that will handle the KISS traffic for the
        asyncio transport.
        """
        return KISSProtocol(
            self._on_connect,
            self._receive,
            self._on_close,
            self._log.getChild('protocol')
        )

    async def _open_connection(self): # pragma: no cover
        """
        Open a connection to the underlying transport.
        """
        raise NotImplementedError('Abstract function')

    def _open(self):
        ensure_future(self._open_connection())

    def _on_connect(self, transport):
        self._transport = transport
        self._init_kiss()

    def _close(self):
        # Wait for all data to be sent.
        self._transport.flush()

        # Close the port
        self._transport.close()

        # Clean up
        self._on_close()

    def _on_close(self, exc=None):
        if exc is not None:
            self._log.error('Closing port due to error %r', exc)

        self._transport = None
        self._state = KISSDeviceState.CLOSED

    def _send_raw_data(self, data):
        self._transport.write(data)


class SerialKISSDevice(BaseTransportDevice):
    """
    A KISS device attached to a serial port.  The serial port may be a
    pseudo-TTY, USB-connected serial port or platform-attached serial port.
    The ``baudrate`` parameter specifies the baud rate used to communicate
    with the TNC, not the speed of the AX.25 network which may be a different
    speed.

    The serial port link is assumed to use 8-bit wide frames, no parity bits
    and one stop bit with no flow control.

    :param device: Device node name to connect to (e.g. `/dev/ttyS0`, `COM3:`)
    :type device: ``str``
    :param baudrate: Baud rate to connect to the serial port at.
    :type baudrate: ``int``
    :Keyword Arguments: These are passed (via ``BaseTransportDevice``) through
                        to ``BaseKISSDevice`` unchanged.
    """
    def __init__(self, device, baudrate, *args, **kwargs):
        super(SerialKISSDevice, self).__init__(*args, **kwargs)
        self._device = device
        self._baudrate = baudrate

    async def _open_connection(self):
        await create_serial_connection(
                self._loop,
                self._make_protocol,
                self._device,
                baudrate=self._baudrate,
                bytesize=EIGHTBITS,
                parity=PARITY_NONE,
                stopbits=STOPBITS_ONE,
                timeout=None, xonxoff=False,
                rtscts=False, write_timeout=None,
                dsrdtr=False, inter_byte_timeout=None
        )


class TCPKISSDevice(BaseTransportDevice):
    """
    A KISS device exposed via a TCP serial server.  This may be a real TNC
    attached to a serial-to-Ethernet gateway, or a soft-TNC.

    :param host: Host name or IP address of the remote TCP server.
    :type device: ``str``
    :param port: Port number on the remote TCP server where the KISS TNC is
                 listening.
    :type port: ``int``
    :param ssl: Whether or not to use Transport Layer Security to connect to
                the remote TCP server.
    :type ssl: ``None``, ``ssl.SSLContext`` or ``True``
    :param family: Socket address family to use when connecting, e.g.
                   ``socket.AF_INET`` for IPv4, ``socket.AF_INET6`` for IPv6,
                   or ``0`` for any.
    :type family: ``int``
    :param proto: Specifies the address protocol.  Exposed for completeness.
                  See the ``socket`` module for further details.
    :type proto: ``int``
    :param flags: Specifies special socket flags used for the connection.  See
                  the ``socket`` module for possible flags.
    :type flags: ``int``
    :param sock: Specifies an optional existing socket object to use for the
                 connection.
    :type sock: ``None`` or ``socket.socket``
    :param local_addr: Local interface address to bind to when connecting.
    :type local_addr: ``None`` or ``str``
    :param server_hostname: Remote server name if needed for Server Name
                            Identification.  In most cases, ``host`` should
                            be a host name and this argument will not be
                            required.  Not used if TLS is disabled.
    :type server_hostname: ``None`` or ``str``
    :Keyword Arguments: These are passed (via ``BaseTransportDevice``) through
                        to ``BaseKISSDevice`` unchanged.
    """
    def __init__(self, host, port, *args, ssl=None, family=0, proto=0, flags=0,
            sock=None, local_addr=None, server_hostname=None, **kwargs):
        super(TCPKISSDevice, self).__init__(*args, **kwargs)

        # Bundle up all the connection arguments together.
        self._conn_args = dict(
                host=host, port=port, ssl=ssl, family=family,
                proto=proto, flags=flags, sock=sock, local_addr=local_addr,
                server_hostname=server_hostname
        )

    async def _open_connection(self):
        await self._loop.create_connection(
                self._make_protocol,
                **self._conn_args
        )


class SubprocKISSDevice(BaseTransportDevice):
    """
    A KISS device that calls a subprocess to communicate with the KISS TNC.
    The subprocess is assumed to accept KISS data on ``stdin`` and emit KISS
    data on ``stdout``.  ``stderr`` traffic is logged but otherwise ignored.

    :param command: Specifies the subprocess command to execute along with any
                    arguments.
    :type command: ``list[str]``
    :param shell: Use a shell to execute the command given.  The command will
                  be concatenated together with spaces to form a single
                  string.  By default, this is ``False``.
    :type shell: ``bool``
    :Keyword Arguments: These are passed (via ``BaseTransportDevice``) through
                        to ``BaseKISSDevice`` unchanged.
    """
    def __init__(self, command, *args, shell=False, **kwargs):
        super(SubprocKISSDevice, self).__init__(*args, **kwargs)
        self._command = command
        self._shell = shell

    def _make_protocol(self):
        """
        Return a SubprocessProtocol instance that will handle the KISS traffic for the
        asyncio transport.
        """
        return KISSSubprocessProtocol(
            self._on_connect,
            self._receive,
            self._on_close,
            self._log.getChild('protocol')
        )

    async def _open_connection(self):
        if self._shell:
            await self._loop.subprocess_shell(
                    self._make_protocol,
                    ' '.join(self._command)
            )
        else:
            await self._loop.subprocess_exec(
                    self._make_protocol,
                    *self._command
            )

    def _send_raw_data(self, data):
        transport = self._transport.get_pipe_transport(0).write(data)


# Port interface


class KISSPort(object):
    """
    A KISS port represents a port interface on a KISS device.  There can be
    a maximum of 16 ports per device, identified by a 4-bit integer.
    """
    def __init__(self, device, port, log):
        """
        Create a new port attached to the given device.
        """
        self._device = device
        self._port = port
        self._log = log

        # Signal for receiving packets
        self.received = Signal()

    @property
    def port(self):
        return self._port

    def send(self, frame):
        """
        Send a raw AX.25 frame to the TNC via this port.
        """
        self._log.debug('XMIT AX.25 %s', frame)
        self._device._send(KISSCmdData(self.port, bytes(frame)))

    def _receive_frame(self, frame):
        """
        Receive and emit an incoming frame from the port.
        """
        self._log.debug('Received frame %s', frame)

        if not isinstance(frame, KISSCmdData):
            # TNC is not supposed to send this!
            return

        self.received.emit(frame=frame.payload)


# Protocol interface adaptors for asyncio


class KISSProtocol(Protocol):
    """
    KISSProtocol basically is a wrapper around asyncio's "Protocol"
    structure.
    """
    def __init__(self, on_connect, on_receive, on_close, log):
        super(KISSProtocol, self).__init__()

        self._on_connect = on_connect
        self._on_receive = on_receive
        self._on_close = on_close
        self._log = log

    def connection_made(self, transport):
        try:
            self._on_connect(transport)
        except:
            self._log.exception('Failed to handle connection establishment')
            transport.close()

    def data_received(self, data):
        try:
            self._on_receive(data)
        except:
            self._log.exception('Failed to handle incoming data')

    def connection_lost(self, exc):
        try:
            self._on_close(exc)
        except:
            self._log.exception('Failed to handle connection loss')


class KISSSubprocessProtocol(Protocol):
    """
    KISSSubprocessProtocol is nearly identical to KISSProtocol but wraps
    SubprocessProtocol instead.
    """
    def __init__(self, on_connect, on_receive, on_close, log):
        super(KISSSubprocessProtocol, self).__init__()

        self._on_connect = on_connect
        self._on_receive = on_receive
        self._on_close = on_close
        self._log = log

    def connection_made(self, transport):
        try:
            self._on_connect(transport)
        except:
            self._log.exception('Failed to handle connection establishment')
            transport.close()

    def pipe_data_received(self, fd, data):
        try:
            if fd == 1: # stdout
                self._on_receive(data)
            else:
                self._log.debug('Data received on fd=%d: %r', fd, data)
        except:
            self._log.exception(
                    'Failed to handle incoming data %r on fd=%d', data, fd
            )

    def process_exited(self):
        try:
            self._on_close(None)
        except:
            self._log.exception('Failed to handle process exit')


# KISS device factory

def make_device(type, **kwargs):
    """
    Create a KISS device of the specified type.  This is a convenience for
    applications that load their configuration via a `dict`-like configuration
    file format such as JSON, YAML or TOML.

    :param type: Type of KISS device to make (see below)
    :type type: ``str``
    :Keyword Arguments: These will be passed to the relevant device class
                        as-is.  Some common arguments for all class types:
      * ``reset_on_close`` (``bool`` = ``True``): Whether or not a "return
        from KISS" command (``C0 FF C0``) should be sent to the TNC on
        closing.
      * ``send_block_size`` (``int`` = ``128``): The number of bytes to send
        in a single write request at a time.  Some KISS TNCs have very small
        serial buffers, and so this, along with ``send_block_delay``, allow
        the outgoing traffic to be "dribbled out" at a rate that avoids
        overflow issues.
      * ``send_block_delay`` (``float`` = ``0.1``): The time to wait between
        consecutive blocks so that the TNC can "catch up".
      * ``kiss_commands`` (``list[str]`` = ``["INT KISS", "RESET"]``):
        The TNC-2 commands to transmit to the TNC after opening to put the
        TNC into KISS mode.  The default value suits Kantronics KPC3 TNCs.
      * ``log`` (``logging.Logger``): A logger interface to log debugging
        traffic.  If not supplied, a default one is created.
      * ``loop`` (``asyncio.AbstractEventLoop``): Asynchronous I/O event loop
        that will schedule the operations for the KISS device.  By default,
        the current event loop (``asyncio.get_event_loop()``) is used.

    +----------------+-------------------------------------------------+
    | ``type`` value | Device type and required arguments              |
    +----------------+-------------------------------------------------+
    | ``serial``     | Serial port KISS device (``SerialKISSDevice``). |
    |                | * ``device`` (``str``):                         |
    |                |   Device name, e.g. `/dev/ttyS0`, `COM3:`       |
    |                | * ``baudrate`` (``int``):                       |
    |                |   Serial port baud rate, e.g. 9600 baud         |
    +----------------+-------------------------------------------------+
    | ``subproc``    | Sub-process KISS device (``SubprocKISSDevice``).|
    |                | * ``command`` (``list[str]``):                  |
    |                |   Command and arguments to execute.             |
    |                | * ``shell`` (``bool``):                         |
    |                |   If set to ``True``, run in a sub-shell.       |
    +----------------+-------------------------------------------------+
    | ``tcp``        | TCP KISS device (``TCPKISSDevice``).            |
    |                | * ``host`` (``str``):                           |
    |                |   IP address or host name of the remote host.   |
    |                | * ``port`` (``int``):                           |
    |                |   TCP port number for the KISS interface.       |
    +----------------+-------------------------------------------------+
    """

    if type == 'serial':
        return SerialKISSDevice(**kwargs)
    elif type == 'subproc':
        return SubprocKISSDevice(**kwargs)
    elif type == 'tcp':
        return TCPKISSDevice(**kwargs)
    else:
        raise ValueError('Unrecognised type=%r' % (type,))



#!/usr/bin/env python3

"""
KISS command unit tests
"""

from aioax25.kiss import KISSPort, KISSCmdData, KISSCommand
from ..nosecompat import eq_
import logging


class DummyKISSDevice(object):
    def __init__(self):
        self.sent = []

    def _send(self, frame):
        self.sent.append(frame)


def test_send():
    """
    Test that frames passed to send are wrapped in a KISS frame and passed up.
    """
    dev = DummyKISSDevice()
    port = KISSPort(dev, 5, logging.getLogger('port'))
    port.send(b'this is a test frame')

    eq_(len(dev.sent), 1)
    last = dev.sent.pop(0)

    assert isinstance(last, KISSCmdData)
    eq_(last.port, 5)
    eq_(last.payload, b'this is a test frame')

def test_receive_frame():
    """
    Test that _receive_frame_recv_frame passes data frame payloads to signal
    """
    sent = []
    dev = DummyKISSDevice()
    port = KISSPort(dev, 5, logging.getLogger('port'))
    port.received.connect(lambda frame, **k : sent.append(frame))

    port._receive_frame(KISSCmdData(port=5, payload=b'this is a test frame'))

    # We should have received that via the signal
    eq_(len(sent),1)
    eq_(sent.pop(0), b'this is a test frame')

def test_receive_frame_filter_nondata():
    """
    Test that _receive_frame filters non-KISSCmdData frames
    """
    sent = []
    dev = DummyKISSDevice()
    port = KISSPort(dev, 5, logging.getLogger('port'))
    port.received.connect(lambda frame, **k : sent.append(frame))

    port._receive_frame(KISSCommand(port=5, cmd=8,
        payload=b'this is a test frame'))

    # We should not have received that frame
    eq_(len(sent),0)



#!/usr/bin/env python3

from ..nosecompat import eq_, assert_set_equal, assert_is, assert_greater, \
        assert_less

from aioax25.aprs.uidigi import APRSDigipeater
from aioax25.frame import AX25UnnumberedInformationFrame, AX25Address
from aioax25.signal import Signal


class DummyAPRSInterface(object):
    def __init__(self):
        self.received_msg = Signal()
        self.mycall = AX25Address.decode('VK4MSL-10')
        self.transmitted = []

    def transmit(self, frame):
        self.transmitted.append(frame)


def test_mydigi_read():
    """
    Test we can obtain a copy of the digipeater call list.
    """
    digipeater = APRSDigipeater()
    digipeater._mydigi.add(AX25Address.decode('VK4MSL'))
    mydigi = digipeater.mydigi

    assert mydigi is not digipeater._mydigi
    eq_(mydigi, digipeater._mydigi)


def test_mydigi_replace():
    """
    Test we can replace the digipeater call list.
    """
    digipeater = APRSDigipeater()
    digipeater._mydigi.add(AX25Address.decode('VK4MSL'))
    digipeater.mydigi = ['VK4BWI']

    eq_(digipeater.mydigi, set([
        AX25Address.decode('VK4BWI')
    ]))


def test_connect_noadd():
    """
    Test we can connect an interface without adding its call to our
    "mydigi" list.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface, addcall=False)
    eq_(digipeater._mydigi, set())


def test_disconnect_norm():
    """
    Test we can disconnect an interface without removing its call from our
    "mydigi" list.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    digipeater.disconnect(interface, rmcall=False)
    eq_(digipeater._mydigi, set([
        AX25Address.decode('VK4MSL-10')
    ]))


def test_rx_irrelevant():
    """
    Test the digipeater module ignores irrelevant frames.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK4RZA',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been dropped
    eq_(len(interface.transmitted), 0)


def test_rx_selfdigied():
    """
    Test the digipeater module ignores frames already digied by us.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK4RZA',
                'VK4MSL-10*',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been dropped
    eq_(len(interface.transmitted), 0)


def test_rx_selftodigi_uplink():
    """
    Test the digipeater module ignores uplink frames with explicit paths.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK4RZA',
                'VK4MSL-10',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been dropped
    eq_(len(interface.transmitted), 0)


def test_rx_selftodigi_first():
    """
    Test the digipeater module digipeats when own call is first.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK4MSL-10',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()

    # It should be passed through VK4MSL-10.
    eq_(str(frame.header.repeaters), 'VK4MSL-10*,VK4RZB')


def test_disconnect():
    """
    Test the digipeater module stops digipeating when disconnected.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK4MSL-10',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()

    # It should be passed through VK4MSL-10.
    eq_(str(frame.header.repeaters), 'VK4MSL-10*,VK4RZB')

    # Disconnect, then send another message
    digipeater.disconnect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK4MSL-10',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing 2'
        )
    )

    # This should not have been digipeated
    eq_(len(interface.transmitted), 0)


def test_rx_selftodigi_alias():
    """
    Test the digipeater module digipeats when alias is first.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.addaliases('GATE')
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'GATE',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()

    # It should be passed through VK4MSL-10.
    eq_(str(frame.header.repeaters), 'VK4MSL-10*,VK4RZB')


def test_rx_selftodigi_downlink():
    """
    Test the digipeater module digipeats when own call is first.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'VK2RXX',   # Ordinarily, this should have the H bit set!
                'VK4RZA*',
                'VK4MSL-10',
                'VK4RZB'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()

    # It should be passed through VK4MSL-10.  H bits of prior repeaters
    # should be left intact.
    eq_(str(frame.header.repeaters), 'VK2RXX,VK4RZA*,VK4MSL-10*,VK4RZB')


def test_rx_exhausted():
    """
    Test the digipeater module ignores WIDEn-0 frames.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'WIDE9-0'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been dropped
    eq_(len(interface.transmitted), 0)


def test_rx_lasthop():
    """
    Test the digipeater module digipeats when we get to the last hop.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'WIDE1-1'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()
    eq_(str(frame.header.repeaters), 'VK4MSL-10*')


def test_rx_nexthop():
    """
    Test the digipeater module appends the correct WIDEn-N next hop.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'WIDE3-3'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()
    eq_(str(frame.header.repeaters), 'VK4MSL-10*,WIDE3-2')


def test_rx_hybridpath():
    """
    Test the digipeater module handles typical APRS paths.
    """
    interface = DummyAPRSInterface()
    digipeater = APRSDigipeater()
    digipeater.connect(interface)
    interface.received_msg.emit(
        interface=interface,
        frame=AX25UnnumberedInformationFrame(
            destination='VK4MSL-1',
            source='VK4MSL-2',
            repeaters=[
                'WIDE1-1',
                'WIDE2-2'
            ],
            pid=0xff, payload=b'testing'
        )
    )

    # This should have been digipeated
    eq_(len(interface.transmitted), 1)
    frame = interface.transmitted.pop()
    eq_(str(frame.header.repeaters), 'VK4MSL-10*,WIDE2-2')



#!/usr/bin/env python3

import logging

import datetime
from aioax25.aprs.datetime import DHMUTCTimestamp, DHMLocalTimestamp, \
        HMSTimestamp, MDHMTimestamp, decode

"""
Time-stamp handling tests.
"""

REF_DT = datetime.datetime(
        2022, 2, 12, 22, 2, 23,
        tzinfo=datetime.timezone.utc
)


def test_dhmutctimestamp_encode():
    """
    Test we can format a DHM timestamp in UTC.
    """
    assert str(DHMUTCTimestamp(12, 21, 52)) == "122152z"


def test_dhmlocaltimestamp_encode():
    """
    Test we can format a DHM timestamp in local time.
    """
    assert str(DHMLocalTimestamp(12, 21, 52)) == "122152/"


def test_hmstimestamp_encode():
    """
    Test we can format a HMS timestamp.
    """
    assert str(HMSTimestamp(21, 59, 6)) == "215906h"


def test_mdhmtimestamp_encode():
    """
    Test we can format a MDHM timestamp.
    """
    assert str(MDHMTimestamp(2, 12, 21, 52)) == "02122152"


def test_dhmutctimestamp_decode():
    """
    Test we can decode a DHM timestamp in UTC.
    """
    res = decode("162152z")
    assert isinstance(res, DHMUTCTimestamp)
    assert (res.day, res.hour, res.minute) == (16, 21, 52)


def test_dhmlocaltimestamp_decode():
    """
    Test we can handle a DHM timestamp in local time.
    """
    res = decode("162152/")
    assert isinstance(res, DHMLocalTimestamp)
    assert (res.day, res.hour, res.minute) == (16, 21, 52)


def test_hmstimestamp_decode():
    """
    Test we can decode a HMS timestamp.
    """
    res = decode("193207h")
    assert isinstance(res, HMSTimestamp)
    assert (res.hour, res.minute, res.second) == (19, 32, 7)


def test_mdhmtimestamp_decode():
    """
    Test we can decode a MDHM timestamp.
    """
    res = decode("02081932")
    assert isinstance(res, MDHMTimestamp)
    assert (res.month, res.day, res.hour, res.minute) \
            == (2, 8, 19, 32)


def test_too_short_decode():
    """
    Test the decoder rejects strings that are too short.
    """
    try:
        decode("09:19")
        assert False, "Should not have decoded"
    except ValueError as e:
        assert str(e) == "Timestamp string too short"


def test_invalid_format():
    """
    Test the decoder rejects strings that are too short.
    """
    try:
        decode("130919\\")
        assert False, "Should not have decoded"
    except ValueError as e:
        assert str(e) == "Timestamp format not recognised"



#!/usr/bin/env python3

from aioax25.frame import AX25Address, AX25FrameHeader
from ..nosecompat import eq_
from ..hex import from_hex, hex_cmp


def test_decode_incomplete():
    """
    Test that an incomplete frame does not cause a crash.
    """
    try:
        AX25FrameHeader.decode(
                from_hex(
                    'ac 96 68 84 ae 92 e0'  # Destination
                )
        )
        assert False, 'This should not have worked'
    except ValueError as e:
        eq_(str(e), 'Too few addresses')

def test_decode_no_digis():
    """
    Test we can decode an AX.25 frame without digipeaters.
    """
    (header, data) = AX25FrameHeader.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'      # Destination
                'ac 96 68 9a a6 98 61'      # Source
            ) + b'frame data goes here'     # Frame data
    )
    eq_(header.destination, AX25Address('VK4BWI', ch=True))
    eq_(header.source, AX25Address('VK4MSL', extension=True))
    eq_(len(header.repeaters), 0)
    eq_(data, b'frame data goes here')

def test_decode_with_1digi():
    """
    Test we can decode an AX.25 frame with one digipeater.
    """
    (header, data) = AX25FrameHeader.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'          # Destination
                'ac 96 68 9a a6 98 60'          # Source
                'ac 96 68 a4 b4 84 61'          # Digi
            ) + b'frame data goes here'         # Frame data
    )
    eq_(header.destination, AX25Address('VK4BWI', ch=True))
    eq_(header.source, AX25Address('VK4MSL'))
    eq_(header.repeaters[0], AX25Address('VK4RZB', extension=True))
    eq_(data, b'frame data goes here')

def test_decode_with_2digis():
    """
    Test we can decode an AX.25 frame with two digipeaters.
    """
    (header, data) = AX25FrameHeader.decode(
            from_hex(
                'ac 96 68 84 ae 92 e0'          # Destination
                'ac 96 68 9a a6 98 60'          # Source
                'ac 96 68 a4 b4 84 60'          # Digi 1
                'ac 96 68 a4 b4 82 61'          # Digi 2
            ) + b'frame data goes here'         # Frame data
    )
    eq_(header.destination, AX25Address('VK4BWI', ch=True))
    eq_(header.source, AX25Address('VK4MSL'))
    eq_(len(header.repeaters), 2)
    eq_(header.repeaters[0], AX25Address('VK4RZB'))
    eq_(header.repeaters[1], AX25Address('VK4RZA', extension=True))
    eq_(data, b'frame data goes here')

def test_encode_no_digis():
    """
    Test we can encode an AX.25 frame without digipeaters.
    """
    header = AX25FrameHeader(destination='VK4BWI', source='VK4MSL', cr=True)
    hex_cmp(bytes(header),
            'ac 96 68 84 ae 92 e0 '         # Destination
            'ac 96 68 9a a6 98 61'          # Source
    )

def test_encode_no_digis_ax25v1():
    """
    Test we can encode an AX.25v1 frame without digipeaters.
    """
    header = AX25FrameHeader(destination='VK4BWI', source='VK4MSL',
            cr=False, src_cr=False)
    hex_cmp(bytes(header),
            'ac 96 68 84 ae 92 60 '         # Destination
            'ac 96 68 9a a6 98 61'          # Source
    )

def test_encode_1digi():
    """
    Test we can encode an AX.25 frame with one digipeater.
    """
    header = AX25FrameHeader(destination='VK4BWI', source='VK4MSL',
            repeaters=('VK4RZB',), cr=True)
    hex_cmp(bytes(header),
            'ac 96 68 84 ae 92 e0 '         # Destination
            'ac 96 68 9a a6 98 60 '         # Source
            'ac 96 68 a4 b4 84 61'          # Digi
    )

def test_encode_2digis():
    """
    Test we can encode an AX.25 frame with two digipeaters.
    """
    header = AX25FrameHeader(destination='VK4BWI', source='VK4MSL',
            repeaters=('VK4RZB', 'VK4RZA'), cr=True)
    hex_cmp(bytes(header),
            'ac 96 68 84 ae 92 e0 '         # Destination
            'ac 96 68 9a a6 98 60 '         # Source
            'ac 96 68 a4 b4 84 60 '         # Digi 1
            'ac 96 68 a4 b4 82 61'          # Digi 2
    )



#!/usr/bin/env python3

"""
APRS symbol handling.
"""

from enum import Enum


PRI_SYMBOL = "/"
SEC_SYMBOL = "\\"
NUM_UNCOMPRESSED = "0123456789"
NUM_COMPRESSED = "abcdefghij"
ALPHA = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"


class APRSSymbolTable(Enum):
    PRIMARY = PRI_SYMBOL
    SECONDARY = SEC_SYMBOL


class APRSOverlayType(Enum):
    NUM_UNCOMPRESSED = "NUM_UNCOMPRESSED"
    NUM_COMPRESSED = "NUM_COMPRESSED"
    ALPHA = "ALPHA"

    @staticmethod
    def identify(overlay):
        try:
            index = NUM_UNCOMPRESSED.index(overlay)
            return (APRSOverlayType.NUM_UNCOMPRESSED, index)
        except ValueError:
            pass

        try:
            index = NUM_COMPRESSED.index(overlay)
            return (APRSOverlayType.NUM_COMPRESSED, index)
        except ValueError:
            pass

        try:
            index = ALPHA.index(overlay)
            return (APRSOverlayType.ALPHA, index)
        except ValueError:
            pass

        raise ValueError("Not a valid overlay character: %r" % overlay)


class APRSSymbol(object):
    """
    Representation of an APRS symbol.
    """
    def __init__(self, table, symbol, overlay=None):
        try:
            table = APRSSymbolTable(table)
        except ValueError:
            # Okay, one of the overlay sets
            overlay = table
            table = APRSSymbolTable.SECONDARY

        # Validate the overlay if given
        if overlay is not None:
            if table != APRSSymbolTable.SECONDARY:
                raise ValueError("Overlays only available on secondary table")
            (overlay_type, overlay) = APRSOverlayType.identify(overlay)
        else:
            overlay_type = None

        self.table = table
        self.symbol = symbol
        self.overlay = overlay
        self.overlay_type = overlay_type

    def __repr__(self): # pragma: no cover
        return (
                '%s(table=%r, symbol=%r, overlay=%r)' \
                    % (
                        self.__class__.__name__,
                        self.table,
                        self.symbol,
                        self.overlay
                    )
        )

    @property
    def tableident(self):
        """
        Return the table identifier character
        """
        if self.overlay_type == APRSOverlayType.NUM_UNCOMPRESSED:
            return NUM_UNCOMPRESSED[self.overlay]
        elif self.overlay_type == APRSOverlayType.NUM_COMPRESSED:
            return NUM_COMPRESSED[self.overlay]
        elif self.overlay_type == APRSOverlayType.ALPHA:
            return ALPHA[self.overlay]
        else:
            return self.table.value



#!/usr/bin/env python3
# These are taken from nosetests
# https://github.com/nose-devs/nose/blob/master/nose/tools/trivial.py
# New tests and code should *NOT* call this module, rather, we should
# migrate existing tests to using `assert` directly, as is py.test style!


"""Tools so trivial that tracebacks should not descend into them
We define the ``__unittest`` symbol in their module namespace so unittest will
skip them when printing tracebacks, just as it does for their corresponding
methods in ``unittest`` proper.
"""
import re
import unittest


__all__ = ['ok_', 'eq_']

# Use the same flag as unittest itself to prevent descent into these functions:
__unittest = 1


def ok_(expr, msg=None):
    """Shorthand for assert. Saves 3 whole characters!
    """
    if not expr:
        raise AssertionError(msg)


def eq_(a, b, msg=None):
    """Shorthand for 'assert a == b, "%r != %r" % (a, b)
    """
    if not a == b:
        raise AssertionError(msg or "%r != %r" % (a, b))


#
# Expose assert* from unittest.TestCase
# - give them pep8 style names
#
caps = re.compile('([A-Z])')

def pep8(name):
    return caps.sub(lambda m: '_' + m.groups()[0].lower(), name)

class Dummy(unittest.TestCase):
    def nop():
        pass
_t = Dummy('nop')

for at in [ at for at in dir(_t)
            if at.startswith('assert') and not '_' in at ]:
    pepd = pep8(at)
    vars()[pepd] = getattr(_t, at)
    __all__.append(pepd)

del Dummy
del _t
del pep8



#!/usr/bin/env python3

"""
Subprocess KISS interface unit tests.
"""

# Most of the functionality here is common to SerialKISSDevice, so this
# really just tests that we pass the right commands to the IOLoop when
# establishing a connection.

from aioax25 import kiss
import logging
from ..asynctest import asynctest
from asyncio import get_event_loop, sleep


@asynctest
async def test_open_connection():
    """
    Test we can open a subprocess without using a shell.
    """
    # This will receive the arguments passed to subprocess_exec
    connection_args = []

    loop = get_event_loop()

    # Stub the subprocess_exec method
    orig_subprocess_exec = loop.subprocess_exec
    async def _subprocess_exec(proto_factory, *args):
        # proto_factory should give us a KISSSubprocessProtocol object
        protocol = proto_factory()
        assert isinstance(protocol, kiss.KISSSubprocessProtocol)

        connection_args.extend(args)
    loop.subprocess_exec = _subprocess_exec

    try:
        device = kiss.SubprocKISSDevice(
                command=['kisspipecmd', 'arg1', 'arg2'],
                loop=loop, log=logging.getLogger(__name__)
        )

        await device._open_connection()

        # Expect a connection attempt to have been made
        assert connection_args == ['kisspipecmd', 'arg1', 'arg2']
    finally:
        # Restore mock
        loop.subprocess_exec = orig_subprocess_exec


@asynctest
async def test_open_connection_shell():
    """
    Test we can open a subprocess using a shell.
    """
    # This will receive the arguments passed to subprocess_shell
    connection_args = []

    loop = get_event_loop()

    # Stub the subprocess_shell method
    orig_subprocess_shell = loop.subprocess_shell
    async def _subprocess_shell(proto_factory, *args):
        # proto_factory should give us a KISSSubprocessProtocol object
        protocol = proto_factory()
        assert isinstance(protocol, kiss.KISSSubprocessProtocol)

        connection_args.extend(args)
    loop.subprocess_shell = _subprocess_shell

    try:
        device = kiss.SubprocKISSDevice(
                command=['kisspipecmd', 'arg1', 'arg2'],
                shell=True,
                loop=loop, log=logging.getLogger(__name__)
        )

        await device._open_connection()

        # Expect a connection attempt to have been made
        assert connection_args == ['kisspipecmd arg1 arg2']
    finally:
        # Restore mock
        loop.subprocess_shell = orig_subprocess_shell


def test_send_raw_data():
    """
    Test data written to the device gets written to the subprocess ``stdin``.
    """
    # This will receive the arguments passed to subprocess_shell
    connection_args = []

    # Mock transport
    class DummyStream(object):
        def __init__(self):
            self.written = b''

        def write(self, data):
            self.written += bytes(data)


    class DummyTransport(object):
        def __init__(self):
            self.stdin = DummyStream()

        def get_pipe_transport(self, fd):
            assert fd == 0
            return self.stdin


    loop = get_event_loop()

    device = kiss.SubprocKISSDevice(
            command=['kisspipecmd', 'arg1', 'arg2'],
            loop=loop, log=logging.getLogger(__name__)
    )
    device._transport = DummyTransport()

    device._send_raw_data(b'testing')

    assert device._transport.stdin.written == b'testing'



#!/usr/bin/env python3

import logging
import gc

from ..nosecompat import eq_, assert_greater, assert_is, \
        assert_is_not, assert_set_equal

from aioax25.aprs.message import APRSMessageHandler, \
        APRSMessageAckFrame, APRSMessageRejFrame, APRSMessageFrame
from aioax25.frame import AX25Address

from ..loop import DummyLoop
from ..hex import to_hex


class DummyAPRSHandler(object):
    def __init__(self):
        self._loop = DummyLoop()
        self._retransmit_count = 2
        self._retransmit_timeout_base = 5
        self._retransmit_timeout_rand = 5
        self._retransmit_timeout_scale = 1.5
        self.mycall = AX25Address.decode('N0CALL')
        self._next_msgid = 12345

        self.sent = []
        self.finished = []
    
    def transmit(self, frame):
        self.sent.append(frame)

    def _on_msg_handler_finish(self, msgid):
        self.finished.append(msgid)


def test_msghandler_addressee():
    """
    Test the handler passes through the addressee given.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    eq_(msghandler.addressee, AX25Address.decode('CQ'))

def test_msghandler_enter_state_success():
    """
    Test the message considers 'SUCCESS' an exit state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the success state.
    msghandler._enter_state(msghandler.HandlerState.SUCCESS)

    # 'done' signal should have been called.
    eq_(len(calls),1)
    call = calls.pop(0)

    assert_set_equal(set(call.keys()), set(['handler', 'state']))
    assert_is(call['handler'], msghandler)
    eq_(call['state'], msghandler.HandlerState.SUCCESS)

def test_msghandler_enter_state_reject():
    """
    Test the message considers 'REJECT' an exit state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the reject state.
    msghandler._enter_state(msghandler.HandlerState.REJECT)

    # 'done' signal should have been called.
    eq_(len(calls),1)
    call = calls.pop(0)

    assert_set_equal(set(call.keys()), set(['handler', 'state']))
    assert_is(call['handler'], msghandler)
    eq_(call['state'], msghandler.HandlerState.REJECT)

def test_msghandler_enter_state_timeout():
    """
    Test the message considers 'TIMEOUT' an exit state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the timeout state.
    msghandler._enter_state(msghandler.HandlerState.TIMEOUT)

    # 'done' signal should have been called.
    eq_(len(calls),1)
    call = calls.pop(0)

    assert_set_equal(set(call.keys()), set(['handler', 'state']))
    assert_is(call['handler'], msghandler)
    eq_(call['state'], msghandler.HandlerState.TIMEOUT)

def test_msghandler_enter_state_cancel():
    """
    Test the message considers 'CANCEL' an exit state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the cancel state.
    msghandler._enter_state(msghandler.HandlerState.CANCEL)

    # 'done' signal should have been called.
    eq_(len(calls),1)
    call = calls.pop(0)

    assert_set_equal(set(call.keys()), set(['handler', 'state']))
    assert_is(call['handler'], msghandler)
    eq_(call['state'], msghandler.HandlerState.CANCEL)

def test_msghandler_enter_state_no_handler():
    """
    Test the message handler can handle the APRS handler disappearing.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    del aprshandler
    gc.collect()

    # Tell it to go to the cancel state.
    msghandler._enter_state(msghandler.HandlerState.CANCEL)

    # 'done' signal should have been called.
    eq_(len(calls),1)
    call = calls.pop(0)

    assert_set_equal(set(call.keys()), set(['handler', 'state']))
    assert_is(call['handler'], msghandler)
    eq_(call['state'], msghandler.HandlerState.CANCEL)

def test_msghandler_enter_state_reject():
    """
    Test the message considers 'REJECT' an exit state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the success state.
    msghandler._enter_state(msghandler.HandlerState.REJECT)

    # 'done' signal should have been called.
    eq_(len(calls),1)
    call = calls.pop(0)

    assert_set_equal(set(call.keys()), set(['handler', 'state']))
    assert_is(call['handler'], msghandler)
    eq_(call['state'], msghandler.HandlerState.REJECT)

def test_msghandler_enter_state_send():
    """
    Test the message considers 'SEND' a intermediate state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the success state.
    msghandler._enter_state(msghandler.HandlerState.SEND)

    # 'done' signal should not have been called.
    eq_(len(calls),0)

    # State should be reflected in the properties
    eq_(msghandler.state, msghandler.HandlerState.SEND)

def test_msghandler_enter_state_retry():
    """
    Test the message considers 'RETRY' a intermediate state.
    """
    calls = []
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))
    msghandler.done.connect(lambda **k : calls.append(k))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Tell it to go to the success state.
    msghandler._enter_state(msghandler.HandlerState.RETRY)

    # 'done' signal should not have been called.
    eq_(len(calls),0)

    # State should be reflected in the properties
    eq_(msghandler.state, msghandler.HandlerState.RETRY)


def test_msghandler_abort_on_no_aprshandler():
    """
    Test the message handler aborts if the parent APRS handler disappears.
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Blow away the APRS handler
    del aprshandler

    # Force garbage collection for pypy's sake.
    gc.collect()

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Now fire the _send method
    msghandler._send()

    # We should have stopped here.
    eq_(msghandler.state, msghandler.HandlerState.FAIL)


def test_msghandler_first_send():
    """
    Test the message handler transmits message on first _send call.
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Message handler is still in the INIT state
    eq_(msghandler.state, msghandler.HandlerState.INIT)

    # Now fire the _send method
    msghandler._send()

    # The message should now be in the 'SEND' state
    eq_(msghandler.state, msghandler.HandlerState.SEND)

    # There should be a pending time-out recorded
    eq_(len(aprshandler._loop.calls), 1)
    (calltime, callfunc) = aprshandler._loop.calls.pop(0)

    # Should be at least 5 seconds from now, calling _on_timeout
    assert_greater(calltime, aprshandler._loop.time() + 5.0)
    eq_(callfunc, msghandler._on_timeout)

def test_msghandler_subsequent_send():
    """
    Test the message handler re-transmits message on subsequent _send calls.
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Force handler into SEND state
    msghandler._state = msghandler.HandlerState.SEND

    # Now fire the _send method
    msghandler._send()

    # The message should now be in the 'RETRY' state
    eq_(msghandler.state, msghandler.HandlerState.RETRY)

    # There should be a pending time-out recorded
    eq_(len(aprshandler._loop.calls), 1)
    (calltime, callfunc) = aprshandler._loop.calls.pop(0)

    # Should be at least 5 seconds from now, calling _on_timeout
    assert_greater(calltime, aprshandler._loop.time() + 5.0)
    eq_(callfunc, msghandler._on_timeout)

    # Retransmit counter should have decremented
    eq_(msghandler._retransmit_count, 1)

def test_msghandler_timeout():
    """
    Test the message handler enters TIMEOUT state when retry count exhausted.
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Force handler into RETRY state
    msghandler._state = msghandler.HandlerState.RETRY

    # Force retransmit count to zero
    msghandler._retransmit_count = 0

    # Now fire the _send method
    msghandler._send()

    # The message should now be in the 'TIMEOUT' state
    eq_(msghandler.state, msghandler.HandlerState.TIMEOUT)

    # There should be no calls pending
    eq_(len(aprshandler._loop.calls), 0)

def test_msghandler_send_invalid_state():
    """
    Test the message handler refuses to send in the wrong state.
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Force handler into TIMEOUT state
    msghandler._state = msghandler.HandlerState.TIMEOUT

    try:
        # Now fire the _send method
        msghandler._send()
        assert False, 'This should have raised a RuntimeError'
    except RuntimeError as e:
        eq_(str(e), 'Incorrect state HandlerState.TIMEOUT')

    # There should be no calls pending
    eq_(len(aprshandler._loop.calls), 0)

def test_msghandler_cancel():
    """
    Test calling cancel stops the timer and enters the CANCEL state.
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Inject a dummy time-out object
    timeout = aprshandler._loop.call_later(1.0, None)
    msghandler._retransmit_timeout = timeout

    # Cancel the message
    msghandler.cancel()

    # Handler should now be in the CANCEL state
    eq_(msghandler.state, msghandler.HandlerState.CANCEL)

    # Our time-out should have been cancelled
    eq_(timeout.cancelled(), True)

    # We should no longer be referencing the time-out
    eq_(msghandler._retransmit_timeout, None)

def test_on_timeout():
    """
    Test calling _on_timeout triggers _send
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    msghandler._on_timeout()

    # There should be a pending _send recorded
    eq_(len(aprshandler._loop.calls), 1)
    (calltime, callfunc) = aprshandler._loop.calls.pop(0)

    # Should be pretty much now, calling _send
    assert_greater(calltime, aprshandler._loop.time() - 0.01)
    eq_(callfunc, msghandler._send)

def test_on_response_timedout():
    """
    Test calling _on_response when in TIMEOUT ignores message
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Force state, suppose we already received a reply, and a well-meaning
    # digi has repeated it.
    msghandler._state = msghandler.HandlerState.SUCCESS
    frame1 = APRSMessageAckFrame(
            destination='APZAIO',
            source='VK4MSL-9',
            addressee='N0CALL',
            msgid='123'
    )
    msghandler._response = frame1

    frame2 = APRSMessageAckFrame(
                destination='APZAIO',
                source='VK4MSL-9',
                addressee='N0CALL',
                msgid='123'
            )
    msghandler._on_response(frame2)

    # Our official response should be the first one
    assert_is(msghandler.response, frame1)
    assert_is_not(msghandler.response, frame2)

def test_on_response_ack():
    """
    Test calling _on_response with ACK when in SEND triggers SUCCESS
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Force state, suppose we just sent our request.
    msghandler._state = msghandler.HandlerState.SEND

    # Pass in our frame
    frame = APRSMessageAckFrame(
            destination='APZAIO',
            source='VK4MSL-9',
            addressee='N0CALL',
            msgid='123'
    )
    msghandler._on_response(frame)

    # Our official response should be the frame we just received
    assert_is(msghandler.response, frame)

    # And we should be done
    eq_(msghandler.state, msghandler.HandlerState.SUCCESS)

def test_on_response_rej():
    """
    Test calling _on_response with REJ when in SEND triggers REJECT
    """
    aprshandler = DummyAPRSHandler()
    msghandler  = APRSMessageHandler(
            aprshandler=aprshandler,
            addressee='CQ',
            path=['WIDE1-1','WIDE2-1'],
            message='testing',
            replyack=False,
            log=logging.getLogger('messagehandler'))

    # Force state, suppose we just sent our request.
    msghandler._state = msghandler.HandlerState.SEND

    # Pass in our frame
    frame = APRSMessageRejFrame(
            destination='APZAIO',
            source='VK4MSL-9',
            addressee='N0CALL',
            msgid='123'
    )
    msghandler._on_response(frame)

    # Our official response should be the frame we just received
    assert_is(msghandler.response, frame)

    # And we should be done
    eq_(msghandler.state, msghandler.HandlerState.REJECT)

def test_message_frame_malformed_start():
    """
    Test the message frame decoder will reject malformed start of message.
    """
    try:
        APRSMessageFrame.decode(None, 'x123456789:This is not valid', None)
    except ValueError as e:
        eq_(str(e), "Not a message frame: 'x123456789:This is not valid'")

def test_message_frame_malformed_delim():
    """
    Test the message frame decoder will reject malformed message delimiter
    """
    try:
        APRSMessageFrame.decode(None, ':123456789xThis is not valid', None)
    except ValueError as e:
        eq_(str(e), "Not a message frame: ':123456789xThis is not valid'")

def test_message_frame_bad_msgid():
    """
    Test the message frame constructor rejects too-big message IDs
    """
    try:
        APRSMessageFrame(
                destination='APRS',
                source='VK4MSL',
                addressee='BREAK',
                message='Break this!',
                msgid=123456
        )
    except ValueError as e:
        eq_(str(e), "message ID '123456' too long")

def test_message_frame_get_msg():
    """
    Test the message frame will return the message enclosed
    """
    msg = APRSMessageFrame(
                destination='APRS',
                source='VK4MSL',
                addressee='TEST',
                message='Station under test',
                msgid=12345
        )
    eq_(msg.message, 'Station under test')

def test_message_frame_copy():
    """
    Test we can copy a message frame
    """
    msg = APRSMessageFrame(
                destination='APRS',
                source='VK4MSL',
                addressee='TEST',
                message='Station under test',
                msgid=12345
        )
    msgcopy = msg.copy()
    assert msg is not msgcopy

    eq_(
            to_hex(bytes(msgcopy)),
            to_hex(bytes(msg))
    )

def test_message_ack_copy():
    """
    Test we can copy a message ACK frame
    """
    msg = APRSMessageAckFrame(
                destination='APRS',
                source='VK4MSL',
                addressee='TEST',
                msgid=12345
        )
    msgcopy = msg.copy()
    assert msg is not msgcopy

    eq_(
            to_hex(bytes(msgcopy)),
            to_hex(bytes(msg))
    )

def test_message_rej_copy():
    """
    Test we can copy a message REJ frame
    """
    msg = APRSMessageRejFrame(
                destination='APRS',
                source='VK4MSL',
                addressee='TEST',
                msgid=12345
        )
    msgcopy = msg.copy()
    assert msg is not msgcopy

    eq_(
            to_hex(bytes(msgcopy)),
            to_hex(bytes(msg))
    )


def test_message_encode_replyack_capable():
    """
    Test we can encode a reply-ack flag.
    """
    msg = APRSMessageFrame(
            destination='APRS',
            source='VK4MSL',
            addressee='VK4BWI',
            message='Test announcing "reply-ack" capability',
            msgid='321',
            replyack=True)
    eq_(msg.payload,
            b':VK4BWI   :Test announcing "reply-ack" capability{321}')


def test_message_encode_replyack_reply():
    """
    Test we can encode a reply-ack reply.
    """
    msg = APRSMessageFrame(
            destination='APRS',
            source='VK4MSL',
            addressee='VK4BWI',
            message='Test reply using "reply-ack" capability',
            msgid='321',
            replyack='567')
    eq_(msg.payload,
            b':VK4BWI   :Test reply using "reply-ack" capability{321}567')


def test_message_encode_noreplyack():
    """
    Test we can encode without reply-ack.
    """
    msg = APRSMessageFrame(
            destination='APRS',
            source='VK4MSL',
            addressee='VK4BWI',
            message='Test without "reply-ack" capability',
            msgid='321')
    eq_(msg.payload,
            b':VK4BWI   :Test without "reply-ack" capability{321')



#!/usr/bin/env python3

"""
APRS messaging handler.
"""

import logging
from functools import partial

from ..signal import Signal
from hashlib import sha256

from .router import APRSRouter
from ..frame import AX25Address
from .frame import APRSFrame
from .message import APRSMessageHandler, \
        APRSMessageFrame, APRSMessageAckFrame, APRSMessageRejFrame


class APRSInterface(APRSRouter):
    def __init__(self, ax25int, mycall,
            # Retransmission parameters
            retransmit_count=4, retransmit_timeout_base=30,
            retransmit_timeout_rand=10, retransmit_timeout_scale=1.5,
            # Destination call to use for our traffic
            aprs_destination='APZAIO',
            # Path to use when sending APRS messages
            aprs_path=['WIDE1-1','WIDE2-1'],
            # AX.25 destination SSIDs to listen for.  *NOTE* Setting this will
            # break reception of MIC-e traffic!  The format here used is a `list`
            # of `dict` objects:
            # listen_destinations=[
            #     # APRS 1.0.1 protocol specification page 13
            #     dict(callsign='^AIR',   regex=True,     ssid=None), # Legacy
            #     dict(callsign='^ALL',   regex=True,     ssid=None),
            #     dict(callsign='^AP',    regex=True,     ssid=None),
            #     dict(callsign='BEACON', regex=False,    ssid=None),
            #     dict(callsign='^CQ',    regex=True,     ssid=None),
            #     dict(callsign='^GPS',   regex=True,     ssid=None),
            #     dict(callsign='^DF',    regex=True,     ssid=None),
            #     dict(callsign='^DGPS',  regex=True,     ssid=None),
            #     dict(callsign='^DRILL', regex=True,     ssid=None),
            #     dict(callsign='^ID',    regex=True,     ssid=None),
            #     dict(callsign='^JAVA',  regex=True,     ssid=None),
            #     dict(callsign='^MAIL',  regex=True,     ssid=None),
            #     dict(callsign='^MICE',  regex=True,     ssid=None),
            #     dict(callsign='^QST',   regex=True,     ssid=None),
            #     dict(callsign='^QTH',   regex=True,     ssid=None),
            #     dict(callsign='^RTCM',  regex=True,     ssid=None),
            #     dict(callsign='^SKY',   regex=True,     ssid=None),
            #     dict(callsign='^SPACE', regex=True,     ssid=None),
            #     dict(callsign='^SPC',   regex=True,     ssid=None),
            #     dict(callsign='^SYM',   regex=True,     ssid=None),
            #     dict(callsign='^TEL',   regex=True,     ssid=None),
            #     dict(callsign='^TEST',  regex=True,     ssid=None),
            #     dict(callsign='^TLM',   regex=True,     ssid=None),
            #     dict(callsign='^WX',    regex=True,     ssid=None),
            #     dict(callsign='^ZIP',   regex=True,     ssid=None)  # Legacy
            # ],
            listen_destinations=None,
            # listen_altnets uses the same format as listen_destinations
            # and adds *additional* altnets using the same specification.
            # *NOTE* Setting this will break reception of MIC-e traffic!
            listen_altnets=None,
            # Maximum message ID modulo function
            msgid_modulo=1000,
            # Length of time in seconds before duplicates expire
            deduplication_expiry=28,
            # Logger instance
            log=None):

        super(APRSRouter, self).__init__()
        if log is None:
            log = logging.getLogger(self.__class__.__module__)

        self._log = log

        # Use the same loop as parent AX.25 interface
        self._loop = ax25int._loop

        # Retransmission settings
        self._retransmit_timeout_base = retransmit_timeout_base
        self._retransmit_timeout_rand = retransmit_timeout_rand
        self._retransmit_timeout_scale = retransmit_timeout_scale
        self._retransmit_count = retransmit_count

        # AX.25 set-up
        self._ax25int = ax25int
        self._mycall = AX25Address.decode(mycall).normalised

        if not (listen_destinations or listen_altnets):
            # Bind to receive all traffic.
            # Whilst there's a list of destination calls for conventional APRS,
            # MIC-e derives the destination from the latitude of the station, thus
            # it could be *ANYTHING*.
            self._ax25int.received_msg.connect(self._on_receive)
        else:
            # Bind to receive traffic (direct, standard APRS destinations, and
            # user-defined alt-nets):
            for spec in [
                    dict(callsign=self._mycall.callsign,
                        ssid=self._mycall.ssid, regex=False)
                    ] + (listen_destinations or []) + (listen_altnets or []):
                self._ax25int.bind(self._on_receive, **spec)


        # Message ID counter
        self._msgid = 0

        # Message ID modulo
        self._msgid_modulo = msgid_modulo

        # APRS destination address for broadcast messages
        self._aprs_destination = AX25Address.decode(
                aprs_destination).normalised

        # APRS digi path
        self._aprs_path = [
                AX25Address.decode(call).normalised
                for call in
                aprs_path or []
        ]

        # Pending messages, by addressee and message ID
        self._pending_msg = {}

        # The messages seen in the last `deduplication_expiry` seconds
        self._msg_expiry = {}
        # Length of time to keep message hashes
        self._deduplication_expiry = deduplication_expiry
        # Time-out handle for deduplication
        self._deduplication_timeout = None

        # Received addressed message call-back.  This fires when a message's
        # addressee field matches 'mycall'
        self.received_addressed_msg = Signal()

    @property
    def mycall(self):
        return self._mycall.copy()

    def send_message(self, addressee, message,
            path=None, oneshot=False, replyack=None):
        """
        Send a APRS message to the named addressee.  If replyack is not None,
        send a reply-ack.  If replyack is True, advertise reply-ack capability.
        """
        if path is None:
            self._log.debug('Setting default path for message to %s',
                    addressee)
            path=self._aprs_path

        if oneshot:
            # Forbid replyack when in one-shot mode as we can't encode a
            # reply-ack otherwise.
            if replyack:
                raise ValueError('Cannot send reply-ack in one-shot mode')

            # One-shot mode, just fire and forget!
            self._log.info('Send one-shot to %s: %s',
                    addressee, message)

            self.transmit(APRSMessageFrame(
                    destination=addressee,
                    source=self.mycall,
                    addressee=addressee,
                    message=message,
                    msgid=None,
                    repeaters=path
            ))
            return
        elif replyack:
            # If replyack is a message, we are replying to *that* message.
            self._log.debug('Reply-ACK is %s', replyack)
            if isinstance(replyack, APRSMessageFrame):
                if not replyack.replyack:
                    # Likely station won't support it
                    raise ValueError('replyack is not a reply-ack message')

                replyack = replyack.msgid

        handler = APRSMessageHandler(self, addressee, path, message,
                replyack, self._log.getChild('message'))
        self._pending_msg[handler.msgid] = handler
        handler._send()
        return handler

    def send_response(self, message, ack=True, direct=False):
        """
        Send a ACK (or if ack=False, REJ) to a numbered message.  If direct is
        true, we send the ACK (or REJ) via the path that the message was received
        by (this complies with AX.25 2.0 but does not work with all APRS
        digipeaters).
        """
        if message.msgid is None:
            return

        self._log.debug('Responding to message %s with ack=%s',
                message, ack)

        if direct and (message.header.repeaters is not None):
            response_path = message.header.repeaters.reply
        else:
            response_path = None

        self.send_message(
                addressee=message.header.source.normalised,
                path=response_path,
                message='%s%s' % ('ack' if ack else 'rej', message.msgid),
                oneshot=True
        )

    @classmethod
    def _hash_frame(cls, frame):
        """
        Generate a hash of the given frame.
        """
        framehash = sha256()
        framehash.update(bytes(frame.header.destination))
        framehash.update(bytes(frame.header.source))
        framehash.update(bytes(frame.control))
        framehash.update(bytes(frame.frame_payload))
        return framehash.digest()

    def _test_or_add_frame(self, frame):
        """
        Hash the given frame, and test to see if it has been seen.
        """
        framedigest = self._hash_frame(frame)
        expiry = self._msg_expiry.get(framedigest, 0)
        if expiry > self._loop.time():
            # We've seen this before.
            return True

        self._msg_expiry[framedigest] = \
                self._loop.time() + self._deduplication_expiry
        self._loop.call_soon(self._schedule_dedup_cleanup)
        return False

    def _schedule_dedup_cleanup(self):
        """
        Schedule a clean-up.
        """
        if self._deduplication_timeout is not None:
            self._deduplication_timeout.cancel()
            self._deduplication_timeout = None

        if len(self._msg_expiry) == 0:
            return

        delay = min(self._msg_expiry.values()) - self._loop.time()
        if delay > 0:
            self._deduplication_timeout = \
                    self._loop.call_later(delay, self._dedup_cleanup)
        else:
            self._loop.call_soon(self._dedup_cleanup)

    def _dedup_cleanup(self):
        """
        Clean up the duplicated message cache.
        """
        now = self._loop.time()
        self._deduplication_timeout = None
        for digest, time in list(self._msg_expiry.items()):
            if time < now:
                self._msg_expiry.pop(digest, None)

        self._loop.call_soon(self._schedule_dedup_cleanup)

    def _on_receive_ack(self, msgid, frame):
        handler = self._pending_msg.get(msgid)
        self._log.debug(
                'Response to %r (pending %r), handler %s',
                msgid, list(self._pending_msg.keys()), handler)
        if handler:
            self._loop.call_soon(handler._on_response, frame)
            return True
        else:
            return False

    def _on_receive(self, frame, **kwargs):
        """
        Handle the incoming frame.
        """
        if self._test_or_add_frame(frame):
            self._log.debug('Ignoring duplicate frame: %s', frame)
            return

        try:
            frame = APRSFrame.decode(frame, self._log.getChild('decode'))
            self._log.debug('Processing incoming message %s (type %s)',
                    frame, frame.__class__.__name__)

            # Pass to the super-class handler
            self._loop.call_soon(partial(
                super(APRSInterface, self)._on_receive,
                frame
            ))

            if isinstance(frame, APRSMessageFrame):
                # This is a message, is it for us?
                if frame.addressee == self.mycall:
                    # Is it a response for us?
                    if isinstance(frame, (APRSMessageAckFrame, \
                            APRSMessageRejFrame)):
                        if self._on_receive_ack(frame.msgid, frame):
                            # Dealt with
                            return
                    else:
                        # Is it a reply-ack?
                        if frame.replyack and (frame.replyack is not True):
                            # Hand to the response handler
                            self._on_receive_ack(frame.replyack, frame)

                        # This is a message addressed to us.
                        self._loop.call_soon(partial(
                            self.received_addressed_msg.emit,
                            interface=self, frame=frame
                        ))
                else:
                    self._log.debug('Addressee is %s, mycall %s, not for us',
                            frame.addressee, self.mycall)
        except:
            self._log.exception('Exception occurred emitting signal')

    def transmit(self, frame):
        """
        Send an AX.25 frame.
        """
        self._log.info('Sending %s', frame)
        try:
            self._ax25int.transmit(frame)
            self._test_or_add_frame(frame)
        except:
            self._log.exception('Failed to send %s', frame)

    @property
    def _next_msgid(self):
        """
        Return the next message ID
        """
        self._msgid = (self._msgid + 1) % self._msgid_modulo
        return str(self._msgid)

    def _on_msg_handler_finish(self, msgid):
        self._pending_msg.pop(msgid, None)



#!/usr/bin/env python3

"""
Human-readable hex strings.
"""

from binascii import a2b_hex

def from_hex(hexstr):
    return a2b_hex(hexstr.replace(' ',''))

def to_hex(bytestr):
    return ' '.join([
        '%02x' % byte
        for byte in bytestr
    ])

def hex_cmp(bytestr1, bytestr2):
    if not isinstance(bytestr1, bytes):
        bytestr1 = from_hex(bytestr1)

    if not isinstance(bytestr2, bytes):
        bytestr2 = from_hex(bytestr2)

    assert bytestr1 == bytestr2, \
            'Byte strings do not match:\n'\
            '  1> %s\n'\
            '  2> %s\n'\
            '1^2> %s' % (to_hex(bytestr1),
                    to_hex(bytestr2),
                    ' '.join([
                        (('%02x' % (a ^ b))
                            if a != b
                            else '  ')
                        for (a, b)
                        in zip(bytestr1, bytestr2)
                    ]))



#!/usr/bin/env python3

from setuptools import setup
from os.path import dirname, join
from sys import version_info
from aioax25 import __version__

requirements = [
        'pyserial',
        'signalslot',
        'pyserial_asyncio'
]

packages = [
        'aioax25',
        'aioax25.aprs',
]

setup(
        name='aioax25',
        url='https://github.com/sjlongland/aioax25/',
        version=__version__,
        author='Stuart Longland VK4MSL',
        author_email='me@vk4msl.id.au',
        license='GPL-2.0-or-later',
        packages=packages,
        requires=requirements,
        install_requires=requirements,
        description='Asynchronous AX.25 interface in pure Python using asyncio',
        long_description=open(join(dirname(__file__), 'README.md'),'r').read(),
        long_description_content_type='text/markdown',
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Environment :: No Input/Output (Daemon)',
            'Framework :: AsyncIO',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
            'Operating System :: POSIX',
            'Programming Language :: Python :: 3 :: Only',
            'Topic :: Communications :: Ham Radio'
        ]
)



#!/usr/bin/env python3

from aioax25.aprs.symbol import APRSSymbol, APRSSymbolTable, APRSOverlayType

"""
Symbol handling tests.
"""

def test_sym_primary():
    """
    Test we can identify a primary symbol.
    """
    sym = APRSSymbol("/", "$")
    assert sym.table == APRSSymbolTable.PRIMARY
    assert sym.symbol == "$"
    assert sym.overlay_type is None
    assert sym.overlay is None

def test_sym_secondary():
    """
    Test we can identify a secondary symbol.
    """
    sym = APRSSymbol("\\", "u")
    assert sym.table == APRSSymbolTable.SECONDARY
    assert sym.symbol == "u"
    assert sym.overlay_type is None
    assert sym.overlay is None

def test_sym_secondary_numover_uncompressed():
    """
    Test we can identify a secondary symbol with a numeric overlay.
    (Uncompressed set)
    """
    sym = APRSSymbol("3", "u")
    assert sym.table == APRSSymbolTable.SECONDARY
    assert sym.symbol == "u"
    assert sym.overlay_type == APRSOverlayType.NUM_UNCOMPRESSED
    assert sym.overlay == 3

def test_sym_secondary_numover_compressed():
    """
    Test we can identify a secondary symbol with a numeric overlay.
    (Compressed set)
    """
    sym = APRSSymbol("d", "u")
    assert sym.table == APRSSymbolTable.SECONDARY
    assert sym.symbol == "u"
    assert sym.overlay_type == APRSOverlayType.NUM_COMPRESSED
    assert sym.overlay == 3

def test_sym_secondary_alphaover():
    """
    Test we can identify a secondary symbol with an alphabetic overlay.
    """
    sym = APRSSymbol("D", "u")
    assert sym.table == APRSSymbolTable.SECONDARY
    assert sym.symbol == "u"
    assert sym.overlay_type == APRSOverlayType.ALPHA
    assert sym.overlay == 3

def test_sym_primary_tableident():
    """
    Test primary symbols' table identifies with "/"
    """
    sym = APRSSymbol(APRSSymbolTable.PRIMARY, "$")
    assert sym.tableident == "/"

def test_sym_secondary_tableident():
    """
    Test we can identify a secondary symbol.
    """
    sym = APRSSymbol(APRSSymbolTable.SECONDARY, "u")
    assert sym.tableident == "\\"

def test_sym_secondary_numover_uncompressed_tableident():
    """
    Test uncompressed numeric overlay is reported with digit.
    """
    sym = APRSSymbol(APRSSymbolTable.SECONDARY, "u", "3")
    assert sym.tableident == "3"

def test_sym_secondary_numover_compressed_tableident():
    """
    Test uncompressed numeric overlay is reported with lower letter.
    """
    sym = APRSSymbol(APRSSymbolTable.SECONDARY, "u", "d")
    assert sym.tableident == "d"

def test_sym_secondary_alphaover_tableident():
    """
    Test alphabetic overlay is reported with upper letter.
    """
    sym = APRSSymbol(APRSSymbolTable.SECONDARY, "u", "D")
    assert sym.tableident == "D"

def test_wrong_table():
    """
    Test that symbol forbids overlays with the wrong table.
    """
    try:
        APRSSymbol(APRSSymbolTable.PRIMARY, "u", "D")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Overlays only available on secondary table"

def test_invalid_overlay():
    try:
        APRSSymbol(APRSSymbolTable.SECONDARY, "u", "x")
        assert False, "Should not have worked"
    except ValueError as e:
        assert str(e) == "Not a valid overlay character: 'x'"



#!/usr/bin/env python3

from ..nosecompat import eq_, assert_set_equal, assert_is, assert_greater, \
        assert_less

import logging
from functools import partial
from signalslot import Signal

from aioax25.aprs import APRSInterface
from aioax25.frame import AX25UnnumberedInformationFrame, AX25Address
from aioax25.aprs.message import APRSMessageFrame, APRSMessageHandler

from ..loop import DummyLoop


class DummyAX25Interface(object):
    def __init__(self):
        self._loop = DummyLoop()
        self.bind_calls = []
        self.transmitted = []
        self.received_msg = Signal()

    def bind(self, callback, callsign, ssid=0, regex=False):
        self.bind_calls.append((callback, callsign, ssid, regex))

    def transmit(self, frame):
        self.transmitted.append(frame)


class DummyMessageHandler(object):
    def _on_response(self):
        pass


def test_constructor_log():
    """
    Test the constructor can accept a logger instance.
    """
    ax25int = DummyAX25Interface()
    log = logging.getLogger('aprslog')
    aprsint = APRSInterface(ax25int, 'VK4MSL-10', log=log)
    assert log is aprsint._log

def test_constructor_bind():
    """
    Test the constructor binds to given destination addresses.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10',
            listen_destinations=[
                dict(callsign='^AIR',   regex=True,     ssid=None), # Legacy
                dict(callsign='^ALL',   regex=True,     ssid=None),
                dict(callsign='^AP',    regex=True,     ssid=None),
                dict(callsign='BEACON', regex=False,    ssid=None),
                dict(callsign='^CQ',    regex=True,     ssid=None),
                dict(callsign='^GPS',   regex=True,     ssid=None),
                dict(callsign='^DF',    regex=True,     ssid=None),
                dict(callsign='^DGPS',  regex=True,     ssid=None),
                dict(callsign='^DRILL', regex=True,     ssid=None),
                dict(callsign='^ID',    regex=True,     ssid=None),
                dict(callsign='^JAVA',  regex=True,     ssid=None),
                dict(callsign='^MAIL',  regex=True,     ssid=None),
                dict(callsign='^MICE',  regex=True,     ssid=None),
                dict(callsign='^QST',   regex=True,     ssid=None),
                dict(callsign='^QTH',   regex=True,     ssid=None),
                dict(callsign='^RTCM',  regex=True,     ssid=None),
                dict(callsign='^SKY',   regex=True,     ssid=None),
                dict(callsign='^SPACE', regex=True,     ssid=None),
                dict(callsign='^SPC',   regex=True,     ssid=None),
                dict(callsign='^SYM',   regex=True,     ssid=None),
                dict(callsign='^TEL',   regex=True,     ssid=None),
                dict(callsign='^TEST',  regex=True,     ssid=None),
                dict(callsign='^TLM',   regex=True,     ssid=None),
                dict(callsign='^WX',    regex=True,     ssid=None),
                dict(callsign='^ZIP',   regex=True,     ssid=None)  # Legacy
            ]
    )
    eq_(len(ax25int.bind_calls), 26)

    assert_set_equal(
            set([
                (call, regex, ssid)
                for (cb, call, ssid, regex)
                in ax25int.bind_calls
            ]),
            set([
                # The first bind call should be for the station SSID
                ('VK4MSL',  False,  10),
                # The rest should be the standard APRS ones.
                ('^AIR',    True,   None),
                ('^ALL',    True,   None),
                ('^AP',     True,   None),
                ('BEACON',  False,  None),
                ('^CQ',     True,   None),
                ('^GPS',    True,   None),
                ('^DF',     True,   None),
                ('^DGPS',   True,   None),
                ('^DRILL',  True,   None),
                ('^ID',     True,   None),
                ('^JAVA',   True,   None),
                ('^MAIL',   True,   None),
                ('^MICE',   True,   None),
                ('^QST',    True,   None),
                ('^QTH',    True,   None),
                ('^RTCM',   True,   None),
                ('^SKY',    True,   None),
                ('^SPACE',  True,   None),
                ('^SPC',    True,   None),
                ('^SYM',    True,   None),
                ('^TEL',    True,   None),
                ('^TEST',   True,   None),
                ('^TLM',    True,   None),
                ('^WX',     True,   None),
                ('^ZIP',    True,   None)
            ])
    )

def test_constructor_bind_destinations_and_altnets():
    """
    Test the constructor binds to both given destinations and alt-nets
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(
            ax25int, 'VK4MSL-10',
            listen_destinations=[
                dict(callsign='^AIR',   regex=True,     ssid=None), # Legacy
                dict(callsign='^ALL',   regex=True,     ssid=None),
                dict(callsign='^AP',    regex=True,     ssid=None),
                dict(callsign='BEACON', regex=False,    ssid=None),
                dict(callsign='^CQ',    regex=True,     ssid=None),
                dict(callsign='^GPS',   regex=True,     ssid=None),
                dict(callsign='^DF',    regex=True,     ssid=None),
                dict(callsign='^DGPS',  regex=True,     ssid=None),
                dict(callsign='^DRILL', regex=True,     ssid=None),
                dict(callsign='^ID',    regex=True,     ssid=None),
                dict(callsign='^JAVA',  regex=True,     ssid=None),
                dict(callsign='^MAIL',  regex=True,     ssid=None),
                dict(callsign='^MICE',  regex=True,     ssid=None),
                dict(callsign='^QST',   regex=True,     ssid=None),
                dict(callsign='^QTH',   regex=True,     ssid=None),
                dict(callsign='^RTCM',  regex=True,     ssid=None),
                dict(callsign='^SKY',   regex=True,     ssid=None),
                dict(callsign='^SPACE', regex=True,     ssid=None),
                dict(callsign='^SPC',   regex=True,     ssid=None),
                dict(callsign='^SYM',   regex=True,     ssid=None),
                dict(callsign='^TEL',   regex=True,     ssid=None),
                dict(callsign='^TEST',  regex=True,     ssid=None),
                dict(callsign='^TLM',   regex=True,     ssid=None),
                dict(callsign='^WX',    regex=True,     ssid=None),
                dict(callsign='^ZIP',   regex=True,     ssid=None)  # Legacy
            ],
            listen_altnets=[
                dict(callsign='VK4BWI', regex=False, ssid=None)
            ])
    eq_(len(ax25int.bind_calls), 27)

    assert_set_equal(
            set([
                (call, regex, ssid)
                for (cb, call, ssid, regex)
                in ax25int.bind_calls
            ]),
            set([
                # The first bind call should be for the station SSID
                ('VK4MSL',  False,  10),
                # The rest should be the standard APRS ones.
                ('^AIR',    True,   None),
                ('^ALL',    True,   None),
                ('^AP',     True,   None),
                ('BEACON',  False,  None),
                ('^CQ',     True,   None),
                ('^GPS',    True,   None),
                ('^DF',     True,   None),
                ('^DGPS',   True,   None),
                ('^DRILL',  True,   None),
                ('^ID',     True,   None),
                ('^JAVA',   True,   None),
                ('^MAIL',   True,   None),
                ('^MICE',   True,   None),
                ('^QST',    True,   None),
                ('^QTH',    True,   None),
                ('^RTCM',   True,   None),
                ('^SKY',    True,   None),
                ('^SPACE',  True,   None),
                ('^SPC',    True,   None),
                ('^SYM',    True,   None),
                ('^TEL',    True,   None),
                ('^TEST',   True,   None),
                ('^TLM',    True,   None),
                ('^WX',     True,   None),
                ('^ZIP',    True,   None),
                # Now should be the "alt-nets"
                ('VK4BWI',  False,  None)
            ])
    )


def test_constructor_bind_altnets():
    """
    Test the constructor binds to "alt-nets" only.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(
            ax25int, 'VK4MSL-10',
            listen_altnets=[
                dict(callsign='VK4BWI', regex=False, ssid=None)
            ])
    eq_(len(ax25int.bind_calls), 2)

    assert_set_equal(
            set([
                (call, regex, ssid)
                for (cb, call, ssid, regex)
                in ax25int.bind_calls
            ]),
            set([
                # The first bind call should be for the station SSID
                ('VK4MSL',  False,  10),
                # Now should be the "alt-nets"
                ('VK4BWI',  False,  None)
            ])
    )

def test_constructor_bind_override():
    """
    Test the constructor allows overriding the usual addresses.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10',
            listen_destinations=[
                dict(callsign='APRS', regex=False, ssid=None)
            ])
    eq_(len(ax25int.bind_calls), 2)

    assert_set_equal(
            set([
                (call, regex, ssid)
                for (cb, call, ssid, regex)
                in ax25int.bind_calls
            ]),
            set([
                # The first bind call should be for the station SSID
                ('VK4MSL',  False,  10),
                # The rest should be the ones we gave
                ('APRS',    False,  None)
            ])
    )

def test_transmit_exception():
    """
    Test that transmit swallows exceptions.
    """
    ax25int = DummyAX25Interface()

    # Stub the transmit so it fails
    calls = []
    def stub(*args):
        calls.append(args)
        raise RuntimeError('Oopsie')
    ax25int.transmit = stub

    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.transmit(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123
            )
    )

    # Transmit should have been called
    eq_(len(calls), 1)

def test_send_message_oneshot():
    """
    Test that send_message in one-shot mode generates a message frame.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    res = aprsint.send_message(
            'VK4MDL-7', 'Hi', oneshot=True
    )

    # We don't get a return value
    assert_is(res, None)

    # No message handler should be registered with the interface
    eq_(len(aprsint._pending_msg), 0)

    # The frame is passed to the AX.25 interface
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message frame
    assert isinstance(frame, APRSMessageFrame)

    # There is no pending messages
    eq_(len(aprsint._pending_msg), 0)

def test_send_message_oneshot_replyack():
    """
    Test that send_message in one-shot mode refuses to send reply-ack.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    try:
        aprsint.send_message(
                'VK4MDL-7', 'Hi', oneshot=True,
                replyack='This should be a message, but the code only tests '
                         'that this value is None, which it won\'t be here.'
        )
    except ValueError as e:
        eq_(str(e), 'Cannot send reply-ack in one-shot mode')

def test_send_message_replyack():
    """
    Test that send_message with a replyack message sets replyack.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    replymsg = APRSMessageFrame(
            destination='APRS',
            source='VK4MDL-7',
            addressee='VK4MSL-7',
            message='Hello',
            msgid='123',
            replyack=True
    )
    res = aprsint.send_message(
            'VK4MDL-7', 'Hi', oneshot=False, replyack=replymsg
    )

    # We got back a handler class
    assert isinstance(res, APRSMessageHandler)

    # That message handler should be registered with the interface
    eq_(len(aprsint._pending_msg), 1)
    assert res.msgid in aprsint._pending_msg
    assert_is(aprsint._pending_msg[res.msgid], res)

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message frame
    assert isinstance(frame, APRSMessageFrame)

    # Frame has reply-ACK set
    eq_(frame.replyack, '123')

    # Message handler is in 'SEND' state
    eq_(res.state, APRSMessageHandler.HandlerState.SEND)

def test_send_message_advreplyack():
    """
    Test that send_message with replyack=True message sets replyack.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    res = aprsint.send_message(
            'VK4MDL-7', 'Hi', oneshot=False, replyack=True
    )

    # We got back a handler class
    assert isinstance(res, APRSMessageHandler)

    # That message handler should be registered with the interface
    eq_(len(aprsint._pending_msg), 1)
    assert res.msgid in aprsint._pending_msg
    assert_is(aprsint._pending_msg[res.msgid], res)

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message frame
    assert isinstance(frame, APRSMessageFrame)

    # Frame has reply-ACK set
    eq_(frame.replyack, True)

    # Message handler is in 'SEND' state
    eq_(res.state, APRSMessageHandler.HandlerState.SEND)

def test_send_message_replyack_notreplyack():
    """
    Test that send_message in confirmable mode generates a message handler.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    replymsg = APRSMessageFrame(
            destination='APRS',
            source='VK4MDL-7',
            addressee='VK4MSL-7',
            message='Hello',
            msgid='123',
            replyack=False
    )
    try:
        aprsint.send_message(
                'VK4MDL-7', 'Hi', oneshot=False, replyack=replymsg
        )
    except ValueError as e:
        eq_(str(e), 'replyack is not a reply-ack message')

def test_send_message_confirmable():
    """
    Test that send_message in confirmable mode generates a message handler.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    res = aprsint.send_message(
            'VK4MDL-7', 'Hi', oneshot=False
    )

    # We got back a handler class
    assert isinstance(res, APRSMessageHandler)

    # That message handler should be registered with the interface
    eq_(len(aprsint._pending_msg), 1)
    assert res.msgid in aprsint._pending_msg
    assert_is(aprsint._pending_msg[res.msgid], res)

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message frame
    assert isinstance(frame, APRSMessageFrame)

    # Message handler is in 'SEND' state
    eq_(res.state, APRSMessageHandler.HandlerState.SEND)

def test_send_response_oneshot():
    """
    Test that send_response ignores one-shot messages.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.send_response(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=None
            )
    )

    # Nothing should be sent
    eq_(len(ax25int.transmitted), 0)

def test_send_response_ack():
    """
    Test that send_response with ack=True sends acknowledgement.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.send_response(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123
            ),
            ack=True
    )

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message acknowledgement frame
    assert isinstance(frame, APRSMessageFrame)
    eq_(frame.payload, b':VK4MSL-10:ack123')

def test_send_response_rej():
    """
    Test that send_response with ack=False sends rejection.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    aprsint.send_response(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123
            ),
            ack=False
    )

    # The APRS message handler will have tried sending the message
    eq_(len(ax25int.transmitted), 1)
    frame = ax25int.transmitted.pop(0)

    # Frame is a APRS message rejection frame
    assert isinstance(frame, APRSMessageFrame)
    eq_(frame.payload, b':VK4MSL-10:rej123')

def test_hash_frame_mismatch_dest():
    """
    Test that _hash_frame returns different hashes for mismatching destination
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-3',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    # These should not be the same
    assert hash1 != hash2

def test_hash_frame_mismatch_src():
    """
    Test that _hash_frame returns different hashes for mismatching source
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-11',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    # These should not be the same
    assert hash1 != hash2

def test_hash_frame_mismatch_payload():
    """
    Test that _hash_frame returns different hashes for mismatching source
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing 1',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing 2',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    # These should not be the same
    assert hash1 != hash2

def test_hash_frame():
    """
    Test that _hash_frame returns a consistent result regardless of digipeaters
    """
    hash1 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
            )
    )

    hash2 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['VK4RZB*','WIDE1-1']
            )
    )

    hash3 = APRSInterface._hash_frame(
            APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['VK4RZB*','VK4RZA*']
            )
    )

    # These should all be the same
    eq_(hash1, hash2)
    eq_(hash1, hash3)

def test_test_or_add_frame_first():
    """
    Test that _test_or_add_frame returns False for new traffic
    """
    frame = APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Try it out
    res = aprsint._test_or_add_frame(frame)

    # We should get 'False' as the response
    eq_(res, False)

    # There should be an entry in our hash table.
    eq_(len(aprsint._msg_expiry), 1)

    # The expiry time should be at least 25 seconds.
    assert_greater(aprsint._msg_expiry.get(framedigest, 0),
            ax25int._loop.time() + 25)

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

def test_test_or_add_frame_repeat():
    """
    Test that _test_or_add_frame returns True for un-expired repeats
    """
    frame = APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    # Inject the frame expiry
    expiry_time = aprsint._loop.time() + 1
    aprsint._msg_expiry[framedigest] = expiry_time

    # Try it out
    res = aprsint._test_or_add_frame(frame)

    # We should get 'False' as the response
    eq_(res, True)

    # Expiry should not have changed
    eq_(len(aprsint._msg_expiry), 1)
    eq_(aprsint._msg_expiry[framedigest], expiry_time)

    # Nothing further should be done.
    eq_(len(ax25int._loop.calls), 0)

def test_test_or_add_frame_expired():
    """
    Test that _test_or_add_frame returns False for expired repeats
    """
    frame = APRSMessageFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                addressee='VK4BWI-2',
                message=b'testing',
                msgid=123,
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')
    # Inject the frame expiry
    expiry_time = aprsint._loop.time() - 1
    aprsint._msg_expiry[framedigest] = expiry_time

    # Try it out
    res = aprsint._test_or_add_frame(frame)

    # We should get 'False' as the response
    eq_(res, False)

    # The expiry time should be at least 25 seconds.
    eq_(len(aprsint._msg_expiry), 1)
    assert_greater(aprsint._msg_expiry.get(framedigest, 0),
            ax25int._loop.time() + 25)

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

def test_schedule_dedup_cleanup_no_msg():
    """
    Test _schedule_dedup_cleanup does nothing if no messages
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    aprsint._schedule_dedup_cleanup()

    # A clean-up should not have been scheduled.
    eq_(len(ax25int._loop.calls), 0)

def test_schedule_dedup_cleanup_pending():
    """
    Test _schedule_dedup_cleanup cancels existing pending clean-ups
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a pending clean-up
    deduplication_timeout = aprsint._loop.call_later(1, None)
    aprsint._deduplication_timeout = deduplication_timeout

    # Schedule the next one
    aprsint._schedule_dedup_cleanup()

    # We should now no-longer have a pending clean-up
    assert deduplication_timeout.cancelled()
    assert_is(aprsint._deduplication_timeout, None)

def test_schedule_dedup_cleanup_oldest_future():
    """
    Test _schedule_dedup_cleanup schedules for expiry of oldest message
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a few hashes
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        b'hash1': now + 1,
        b'hash2': now + 2,
        b'hash3': now + 3
    })

    # Schedule the clean-up
    aprsint._schedule_dedup_cleanup()

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (calltime, callfunc) = ax25int._loop.calls.pop(0)

    # Should be scheduled for the earliest expiry
    assert_less(calltime - now, 1.01)
    assert_greater(calltime - now, 0.99)
    eq_(callfunc, aprsint._dedup_cleanup)

def test_schedule_dedup_cleanup_oldest_past():
    """
    Test _schedule_dedup_cleanup schedules immediately for expired messages.
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a few hashes
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        b'hash1': now - 1,
        b'hash2': now + 2,
        b'hash3': now + 3
    })

    # Schedule the clean-up
    aprsint._schedule_dedup_cleanup()

    # A clean-up should have been scheduled.
    eq_(len(ax25int._loop.calls), 1)
    (calltime, callfunc) = ax25int._loop.calls.pop(0)

    # Should be scheduled pretty much now
    assert_less(calltime - now, 0.01)
    eq_(callfunc, aprsint._dedup_cleanup)

def test_dedup_cleanup_expired():
    """
    Test _dedup_cleanup removes only expired messages
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a few hashes
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        b'hash1': now - 1,
        b'hash2': now + 2,
        b'hash3': now + 3
    })

    # Perform the clean-up
    aprsint._dedup_cleanup()

    # We should no longer have 'hash1'
    assert b'hash1' not in aprsint._msg_expiry

    # but should have the others
    assert b'hash2' in aprsint._msg_expiry
    assert b'hash3' in aprsint._msg_expiry

    # There should be a re-schedule pending
    eq_(len(ax25int._loop.calls), 1)
    (calltime, callfunc) = ax25int._loop.calls.pop(0)

    # Should be scheduled pretty much now
    assert_less(calltime - now, 0.01)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

def test_on_receive_exception():
    """
    Test _on_receive swallows exceptions.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b':VK4BWI-2 :testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Stub the _test_or_add_frame so it returns false
    aprsint._test_or_add_frame = lambda *a : False

    # Stub the IOLoop's call_soon so it fails
    calls = []
    def stub(*args):
        calls.append(args)
        raise RuntimeError('Oopsie')
    aprsint._loop.call_soon = stub

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # We should have called call_soon, but the exception should have been
    # caught and logged.
    eq_(len(calls), 1)

def test_on_receive_notframe():
    """
    Test _on_receive ignores non-APRS-frames.
    """
    # Create a frame
    class DummyFrame(AX25UnnumberedInformationFrame):
        def __init__(self, *args, **kwargs):
            self.addressee_calls = 0
            super(DummyFrame, self).__init__(*args, **kwargs)

        @property
        def addressee(self):
            self.addressee_calls += 1
            return AX25Address.decode('N0CALL')


    frame = DummyFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b'this is not an APRS message',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Stub the _test_or_add_frame so it returns false
    aprsint._test_or_add_frame = lambda *a : False

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # The addressee property should not be touched
    eq_(frame.addressee_calls, 0)

def test_on_receive_dup():
    """
    Test _on_receive ignores duplicate frames.
    """
    # Create a frame and hash it
    frame = AX25UnnumberedInformationFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b':VK4BWI-2 :testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )
    framedigest = APRSInterface._hash_frame(frame)

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject the hash
    now = aprsint._loop.time()
    aprsint._msg_expiry.update({
        framedigest: now + 3
    })

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be no calls made
    eq_(len(ax25int._loop.calls), 0)

def test_on_receive_pass_to_router():
    """
    Test _on_receive passes the message to the base APRSRouter class.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='VK4BWI-2',
                source='VK4MSL-10',
                pid=0xf0,
                payload=b':VK4BWI-2 :testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be two calls made, one to our deduplication clean-up, the
    # other to our superclass
    eq_(len(ax25int._loop.calls), 2)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

def test_on_receive_addressed():
    """
    Test _on_receive of message addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:testing{123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be three calls made, one to our deduplication clean-up, the
    # other to our superclass, the third to our received_address_msg signal.
    eq_(len(ax25int._loop.calls), 3)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, aprsint.received_addressed_msg.emit)

def test_on_receive_addressed_replyack():
    """
    Test _on_receive of message addressed to station advertising reply-ack.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:testing{123}',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be three calls made, one to our deduplication clean-up, the
    # other to our superclass, the third to our received_address_msg signal.
    eq_(len(ax25int._loop.calls), 3)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, aprsint.received_addressed_msg.emit)

def test_on_receive_unsol_ackrej():
    """
    Test _on_receive of unsolicited ACK/REJ addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:ack123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be two calls made, one to our deduplication clean-up, the
    # other to our superclass.  We don't pass the message out otherwise.
    eq_(len(ax25int._loop.calls), 2)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

def test_on_receive_sol_ackrej():
    """
    Test _on_receive of solicited ACK/REJ addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:ack123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a message handler for the message ID
    handler = DummyMessageHandler()
    aprsint._pending_msg['123'] = handler

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be three calls made, one to our deduplication clean-up, the
    # second to our superclass, and finally the third to the handler's
    # _on_response method.
    eq_(len(ax25int._loop.calls), 3)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

    (_, callfunc, msg) = ax25int._loop.calls.pop(0)
    eq_(callfunc, handler._on_response)
    eq_(bytes(frame), bytes(msg))

def test_on_receive_sol_replyack():
    """
    Test _on_receive of solicited reply-ack addressed to station.
    """
    # Create a frame
    frame = AX25UnnumberedInformationFrame(
                destination='APZAIO',
                source='VK4BWI-2',
                pid=0xf0,
                payload=b':VK4MSL-10:testing{356}123',
                repeaters=['WIDE2-1','WIDE1-1']
    )

    # Create our interface
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a message handler for the message ID
    handler = DummyMessageHandler()
    aprsint._pending_msg['123'] = handler

    # Now pass the frame in as if it were just received
    aprsint._on_receive(frame)

    # There should be four calls made, one to our deduplication clean-up, the
    # second to our superclass, the third to the handler's _on_response method
    # and finally the incoming message should be emitted like a normal message.
    eq_(len(ax25int._loop.calls), 4)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    eq_(callfunc, aprsint._schedule_dedup_cleanup)

    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, super(APRSInterface, aprsint)._on_receive)

    (_, callfunc, msg) = ax25int._loop.calls.pop(0)
    eq_(callfunc, handler._on_response)
    eq_(bytes(frame), bytes(msg))

    # The message should also have been treated as a new incoming message.
    (_, callfunc) = ax25int._loop.calls.pop(0)
    assert isinstance(callfunc, partial)
    eq_(callfunc.func, aprsint.received_addressed_msg.emit)


def test_on_msg_handler_finish():
    """
    Test that _on_msg_handler_finish removes a message from the pending list
    """
    ax25int = DummyAX25Interface()
    aprsint = APRSInterface(ax25int, 'VK4MSL-10')

    # Inject a message handler for the message ID
    handler = DummyMessageHandler()
    aprsint._pending_msg['123'] = handler

    # Call the clean-up function
    aprsint._on_msg_handler_finish('123')

    # This should now be empty
    eq_(aprsint._pending_msg, {})



#!/usr/bin/env python3

from aioax25.frame import AX25Address
from ..nosecompat import eq_
from ..hex import from_hex, to_hex

def test_decode_wrongtype():
    """
    Test that incorrect types are handled.
    """
    try:
        AX25Address.decode(12345)
        assert False, 'Should not have worked'
    except TypeError as e:
        eq_(str(e), "Don't know how to decode 12345")

def test_decode_bytes_short():
    """
    Test decoding a truncated address does not crash.
    """
    try:
        AX25Address.decode(
                from_hex('ac 96 68 9a a6')
        )
        assert False, 'This should not work'
    except ValueError as e:
        eq_(str(e), 'AX.25 addresses must be 7 bytes!')

def test_decode_bytes():
    """
    Test we can decode a plain AX.25 address in binary.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 9a a6 98 00')
    )
    eq_(addr._callsign, 'VK4MSL')

def test_decode_bytes_spaces():
    """
    Test trailing spaces are truncated in call-signs.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 84 82 40 00')
    )
    eq_(addr._callsign, 'VK4BA')

def test_decode_bytes_ext():
    """
    Test we can decode the extension bit set in binary.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 9a a6 98 01')
    )
    eq_(addr._extension, True)

def test_decode_bytes_ssid():
    """
    Test we can decode the SSID in binary.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 9a a6 98 14')
    )
    eq_(addr._ssid, 10)

def test_decode_bytes_res0():
    """
    Test we can decode the first reserved bit in binary.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 9a a6 98 20')
    )
    eq_(addr._res0, True)

def test_decode_bytes_res1():
    """
    Test we can decode the first reserved bit in binary.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 9a a6 98 40')
    )
    eq_(addr._res1, True)

def test_decode_bytes_ch():
    """
    Test we can decode the C/H bit in binary.
    """
    addr = AX25Address.decode(
            from_hex('ac 96 68 9a a6 98 80')
    )
    eq_(addr._ch, True)

def test_decode_str():
    """
    Test that we can decode a call-sign into an AX.25 address.
    """
    addr = AX25Address.decode(
            'VK4MSL'
    )
    eq_(addr._callsign, 'VK4MSL')

def test_decode_str_invalid():
    """
    Test that strings are correctly validated.
    """
    try:
        AX25Address.decode(
                'VK4-MSL'
        )
        assert False, 'Should not have worked'
    except ValueError as e:
        eq_(str(e), 'Not a valid SSID: VK4-MSL')

def test_decode_str_ssid():
    """
    Test that we can decode the SSID in a string.
    """
    addr = AX25Address.decode(
            'VK4MSL-12'
    )
    eq_(addr._ssid, 12)

def test_decode_str_ch():
    """
    Test that we can decode the C/H bit in a string.
    """
    addr = AX25Address.decode(
            'VK4MSL*'
    )
    eq_(addr._ch, True)

def test_decode_ax25address():
    """
    Test that passing in a AX25Address results in a clone being made.
    """
    addr1 = AX25Address('VK4MSL', 5)
    addr2 = AX25Address.decode(addr1)
    assert addr1 is not addr2
    for field in ('_callsign', '_ssid', '_ch', \
            '_res0', '_res1', '_extension'):
        eq_(getattr(addr1, field), getattr(addr2, field))

def test_encode_str():
    """
    Test we can encode a AX25Address as a string
    """
    eq_(str(AX25Address('VK4MSL', 0)), 'VK4MSL')

def test_encode_str_ssid():
    """
    Test we can encode a AX25Address as a string
    """
    eq_(str(AX25Address('VK4MSL', 11)), 'VK4MSL-11')

def test_encode_str_ch():
    """
    Test we can encode a AX25Address' C/H bit as a string
    """
    eq_(str(AX25Address('VK4MSL', ch=True)), 'VK4MSL*')

def test_encode_repr():
    """
    Test we can represent the AX25Address as a Python string
    """
    eq_(repr(AX25Address('VK4MSL', ch=True)), \
            ('AX25Address(callsign=VK4MSL, ssid=0, ch=True, '\
             'res0=True, res1=True, extension=False)'))

def test_encode_bytes():
    """
    Test we can encode a AX25Address as binary
    """
    eq_(
            to_hex(bytes(AX25Address('VK4MSL', 0,
                res0=False, res1=False, ch=False,
                extension=False))),
            'ac 96 68 9a a6 98 00'
    )

def test_encode_bytes_ssid():
    """
    Test we can encode a AX25Address as binary
    """
    eq_(
            to_hex(bytes(AX25Address('VK4MSL', 11,
                res0=False, res1=False, ch=False,
                extension=False))),
            'ac 96 68 9a a6 98 16'
    )

def test_encode_bytes_ch():
    """
    Test we can encode a AX25Address' C/H bit as binary
    """
    eq_(
            to_hex(bytes(AX25Address('VK4MSL',
                res0=False, res1=False, ch=True,
                extension=False))),
            'ac 96 68 9a a6 98 80'
    )

def test_encode_bytes_ext():
    """
    Test we can encode a AX25Address' extension bit as binary
    """
    eq_(
            to_hex(bytes(AX25Address('VK4MSL',
                res0=False, res1=False, ch=False,
                extension=True))),
            'ac 96 68 9a a6 98 01'
    )

def test_encode_bytes_res1():
    """
    Test we can encode a AX25Address' Reserved 1 bit as binary
    """
    eq_(
            to_hex(bytes(AX25Address('VK4MSL',
                res0=False, res1=True, ch=False,
                extension=False))),
            'ac 96 68 9a a6 98 40'
    )

def test_encode_bytes_res0():
    """
    Test we can encode a AX25Address' Reserved 0 bit as binary
    """
    eq_(
            to_hex(bytes(AX25Address('VK4MSL',
                res0=True, res1=False, ch=False,
                extension=False))),
            'ac 96 68 9a a6 98 20'
    )

def test_eq_match():
    """
    Test the __eq__ operator correctly matches addresses.
    """
    a = AX25Address('VK4MSL', 12, ch=False)
    b = AX25Address('VK4MSL', 12, ch=False)
    assert a is not b
    assert a == b

def test_eq_notmatch():
    """
    Test the __eq__ operator correctly identifies non-matching addresses.
    """
    a = AX25Address('VK4MSL', 12, ch=False)
    b = AX25Address('VK4MSL', 12, ch=True)
    assert a != b

def test_eq_notaddr():
    """
    Test the __eq__ operator does not attempt to compare non-addresses.
    """
    a = AX25Address('VK4MSL', 12, ch=False)
    assert a != 'foobar'

def test_hash():
    """
    Test we can obtain a reliable hash.
    """
    a = AX25Address('VK4MSL', 12, ch=False)
    b = AX25Address('VK4MSL', 12, ch=False)
    c = AX25Address('VK4MSL', 12, ch=True)
    assert a is not b
    assert a is not c
    assert hash(a) == hash(b)
    assert hash(a) != hash(c)

def test_copy():
    """
    Test we can make copies of the address with arbitrary fields set.
    """
    a = AX25Address('VK4MSL', 15, ch=False)
    b = a.copy(ch=True)

    assert b._ch is True

    # Everything else should be the same
    for field in ('_callsign', '_ssid', '_res0', '_res1', '_extension'):
        eq_(getattr(a, field), getattr(b, field))

def test_normalised():
    """
    Test we can get normalised copies for comparison.
    """
    a = AX25Address('VK4MSL', 15, ch=True, res0=False, res1=False)
    b = a.normalised

    assert b._ch is False
    assert b._res0 is True
    assert b._res1 is True

def test_ch_setter():
    """
    Test we can mutate the C/H bit.
    """
    a = AX25Address('VK4MSL', 15, ch=False)
    a.ch = True
    assert a._ch is True

def test_extension_setter():
    """
    Test we can mutate the extension bit.
    """
    a = AX25Address('VK4MSL', 15, extension=False)
    a.extension = True
    assert a._extension is True



#!/usr/bin/env python3

__author__ = 'Stuart Longland VK4MSL'
__email__ = 'me@vk4msl.id.au'
__license__ = 'GPL-2.0-or-later'
__copyright__ = 'Copyright 2021, Stuart Longland (and contributors)'
__version__ = '0.0.11.post0'
